#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.07 22:00:00                  #
# ================================================== #

import os
import time
from typing import List, Tuple

from PySide6.QtCore import QTimer, QObject
from pydub import AudioSegment


class NativeBackend(QObject):

    MIN_FRAMES = 25  # minimum frames to start transcription
    AUTO_CONVERT_TO_WAV = False  # automatically convert to WAV format

    def __init__(self, window=None):
        """
        Audio input capture core

        :param window: Window instance
        """
        super().__init__(window)
        self.window = window
        self.audio_source = None
        self.audio_io_device = None
        self.frames = []
        self.actual_audio_format = None
        self.path = None
        self.disconnected = False
        self.initialized = False
        self.start_time = 0
        self.devices = []
        self.selected_device = None
        self.bar = None
        self.loop = False
        self.stop_callback = None
        self.player = None
        self.playback_timer = None
        self.volume_timer = None
        self.audio_output = None
        self.envelope = []
        self.chunk_ms = 10

        # Get configuration values (use defaults if unavailable)
        if self.window is not None and hasattr(self.window, "core"):
            self.channels = int(self.window.core.config.get('audio.input.channels', 1))
            self.rate = int(self.window.core.config.get('audio.input.rate', 44100))
        else:
            self.channels = 1
            self.rate = 44100

    def init(self):
        """
        Initialize audio input backend.
        """
        if not self.initialized:
            self.check_audio_devices()
            self.initialized = True

    def set_repeat_callback(self, callback):
        """
        Set callback to be called on loop recording

        :param callback: function to call on loop recording
        """
        if callable(callback):
            self.stop_callback = callback
        else:
            raise ValueError("Callback must be a callable function")

    def set_loop(self, loop: bool):
        """
        Set loop recording

        :param loop: True to enable loop recording
        """
        self.loop = loop

    def set_path(self, path: str):
        """
        Set audio input file path

        :param path: file path
        """
        self.path = path

    def set_bar(self, bar):
        """
        Set audio level bar

        :param bar: audio level bar
        """
        self.bar = bar

    def start(self):
        """
        Start audio input recording

        :return: True if started
        """
        self.init()
        # Clear previous frames
        self.frames = []

        # Prepare selected device
        self.prepare_device()
        if not self.selected_device:
            print("No audio input device selected")
            return
        if self.disconnected:
            print("Audio source disconnected, please connect the audio source")
            return False

        # Prevent multiple recordings
        if self.audio_source is not None:
            return False

        # Set up audio input and start recording
        self.setup_audio_input()
        self.start_time = time.time()
        return True

    def stop(self) -> bool:
        """
        Stop audio input recording

        :return: True if stopped
        """
        result = False
        if self.audio_source is not None:
            # Disconnect the readyRead signal
            self.audio_io_device.readyRead.disconnect(self.process_audio_input)
            self.audio_source.stop()
            self.audio_source = None
            self.audio_io_device = None

            # Save frames to file (if any)
            if self.frames:
                self.save_audio_file(self.path)
                result = True
            else:
                print("No audio data recorded")
        return result

    def has_source(self) -> bool:
        """
        Check if audio source is available

        :return: True if available
        """
        if self.audio_source is not None:
            return True
        return False

    def has_frames(self) -> bool:
        """
        Check if audio frames are available

        :return: True if available
        """
        if self.frames:
            return True
        return False

    def has_min_frames(self) -> bool:
        """
        Check if min required audio frames

        :return: True if min frames
        """
        if self.frames:
            frames = self.get_frames()
            if len(frames) < self.MIN_FRAMES:
                return False
            else:
                return True
        return False

    def reset_audio_level(self):
        """Reset the audio level bar"""
        if self.bar is not None:
            self.bar.setLevel(0)

    def check_audio_input(self) -> bool:
        """
        Check if default audio input device is working using native PySide.

        :return: True if working
        """
        self.init()
        from PySide6.QtMultimedia import QMediaDevices, QAudioFormat, QAudioSource

        devices = QMediaDevices.audioInputs()
        if not devices:
            print("No audio input devices found.")
            return False

        device = self.selected_device if self.selected_device else devices[0]
        audio_format = QAudioFormat()
        audio_format.setSampleRate(self.rate)
        audio_format.setChannelCount(self.channels)
        audio_format.setSampleFormat(QAudioFormat.SampleFormat.Int16)

        if not device.isFormatSupported(audio_format):
            print("Requested format not supported, using nearest format.")
            audio_format = device.preferredFormat()
        try:
            audio_source = QAudioSource(device, audio_format)
            io_device = audio_source.start()
            if io_device is None:
                print("Unable to access audio input device")
                return False
            audio_source.stop()
            return True
        except Exception as e:
            print(f"Error while checking audio input: {e}")
            return False

    def check_audio_devices(self):
        """Check audio input devices"""
        from PySide6.QtMultimedia import QMediaDevices
        self.devices = QMediaDevices.audioInputs()
        if not self.devices:
            # no devices found
            self.selected_device = None
            print("No audio input devices found.")
        else:
            current = int(self.window.core.config.get('audio.input.device', 0))
            if current < 0 or current >= len(self.devices):
                # if current device is not valid, set the first device as default
                print(f"Invalid audio input device index {current}, using the first device.")
                current = 0
            self.selected_device = self.devices[current]

    def device_changed(self, index: int):
        """
        Change audio input device

        :param index: device index
        """
        self.init()
        if 0 <= index < len(self.devices):
            self.selected_device = self.devices[index]
        else:
            self.selected_device = None

    def prepare_device(self):
        """Set the current audio input device"""
        self.init()
        device_id = int(self.window.core.config.get('audio.input.device', 0))
        self.device_changed(device_id)

    def get_frames(self) -> list:
        """
        Get recorded audio frames

        :return: list of frames
        """
        return self.frames

    def setup_audio_input(self):
        """Set up audio input device and start recording"""
        self.init()
        if not self.selected_device:
            print("No audio input device selected")
            return

        from PySide6.QtMultimedia import QAudioFormat, QAudioSource

        # Define audio format
        audio_format = QAudioFormat()
        audio_format.setSampleRate(int(self.window.core.config.get('audio.input.rate', 44100)))
        audio_format.setChannelCount(int(self.window.core.config.get('audio.input.channels', 1)))
        audio_format.setSampleFormat(QAudioFormat.SampleFormat.Int16)

        # Select default audio input device
        audio_input_device = self.selected_device

        # Check if the format is supported
        if not audio_input_device.isFormatSupported(audio_format):
            print("Requested format not supported, using nearest format.")
            audio_format = audio_input_device.preferredFormat()
            if audio_format.channelCount() > 2:
                audio_format.setChannelCount(2)
            if audio_format.sampleRate() > 44100:
                audio_format.setSampleRate(44100)
            if audio_format.bytesPerSample() > 2:
                audio_format.setSampleFormat(QAudioFormat.SampleFormat.Int16)

        # Store the actual format being used
        self.actual_audio_format = audio_format

        # Create QAudioSource object with the device and format
        try:
            self.audio_source = QAudioSource(audio_input_device, audio_format)
        except Exception as e:
            self.disconnected = True
            print(f"Failed to create audio source: {e}")

        # Start audio input and obtain the QIODevice
        try:
            self.audio_io_device = self.audio_source.start()
            if self.audio_io_device is None:
                raise Exception("Unable to access audio input device")
        except Exception as e:
            print(f"Failed to start audio input: {e}")
            self.disconnected = True
            self.audio_source = None
            self.audio_io_device = None
            return

        # Connect the readyRead signal to process incoming data
        self.audio_io_device.readyRead.connect(self.process_audio_input)

    def process_audio_input(self):
        """Process incoming audio data"""
        import numpy as np
        from PySide6.QtMultimedia import QAudioFormat

        # add seconds to stop timer
        data = self.audio_io_device.readAll()
        if data.isEmpty():
            return

        # Convert QByteArray to bytes
        data_bytes = data.data()

        # Append raw data to frames list for saving
        self.frames.append(data_bytes)

        # Determine the correct dtype and normalization factor
        sample_format = self.actual_audio_format.sampleFormat()
        dtype = self.get_dtype_from_sample_format(sample_format)
        normalization_factor = self.get_normalization_factor(sample_format)

        # Convert bytes to NumPy array of the appropriate type
        samples = np.frombuffer(data_bytes, dtype=dtype)
        if samples.size == 0:
            return

        # For unsigned formats, center the data
        if sample_format == QAudioFormat.SampleFormat.UInt8:
            samples = samples.astype(np.int16)
            samples -= 128

        # Compute RMS of the audio samples as float64 for precision
        rms = np.sqrt(np.mean(samples.astype(np.float64) ** 2))

        # Normalize RMS value based on the sample format
        level = rms / normalization_factor

        # Ensure level is within 0.0 to 1.0
        level = min(max(level, 0.0), 1.0)

        # Scale to 0-100
        level_percent = level * 100

        # Update the level bar widget
        self.update_audio_level(level_percent)

        # Handle loop recording
        if self.loop and self.stop_callback is not None:
            stop_interval = int(self.window.core.config.get('audio.input.stop_interval', 10))
            current_time = time.time()
            time_elapsed = current_time - self.start_time
            if time_elapsed >= stop_interval:
                # stop recording here, save audio chunk to WAV file, run transcription, and start recording again
                self.start_time = current_time
                QTimer.singleShot(0, self.stop_callback)  # required QTimer to prevent crash!!!

    def update_audio_level(self, level: int):
        """
        Update the audio level bar

        :param level: audio level
        """
        if self.bar is not None:
            self.bar.setLevel(level)

    def save_audio_file(self, filename: str):
        """
        Save the recorded audio frames to a WAV file

        :param filename: output file name
        """
        import wave
        # Define the parameters for the WAV file
        channels = self.actual_audio_format.channelCount()
        sample_size = self.actual_audio_format.bytesPerSample()
        frame_rate = self.actual_audio_format.sampleRate()

        # Open the WAV file
        wf = wave.open(filename, 'wb')
        wf.setnchannels(channels)
        wf.setsampwidth(sample_size)
        wf.setframerate(frame_rate)

        # Write frames to the file
        wf.writeframes(b''.join(self.frames))
        wf.close()

    def get_dtype_from_sample_format(self, sample_format):
        """
        Get the NumPy dtype corresponding to the QAudioFormat sample format

        :param sample_format: QAudioFormat.SampleFormat
        """
        import numpy as np
        from PySide6.QtMultimedia import QAudioFormat

        if sample_format == QAudioFormat.SampleFormat.UInt8:
            return np.uint8
        elif sample_format == QAudioFormat.SampleFormat.Int16:
            return np.int16
        elif sample_format == QAudioFormat.SampleFormat.Int32:
            return np.int32
        elif sample_format == QAudioFormat.SampleFormat.Float:
            return np.float32
        else:
            raise ValueError("Unsupported sample format")

    def get_normalization_factor(self, sample_format):
        """
        Get the normalization factor for the QAudioFormat sample format

        :param sample_format: QAudioFormat.SampleFormat
        """
        from PySide6.QtMultimedia import QAudioFormat

        if sample_format == QAudioFormat.SampleFormat.UInt8:
            return 255.0
        elif sample_format == QAudioFormat.SampleFormat.Int16:
            return 32768.0
        elif sample_format == QAudioFormat.SampleFormat.Int32:
            return float(2 ** 31)
        elif sample_format == QAudioFormat.SampleFormat.Float:
            return 1.0
        else:
            raise ValueError("Unsupported sample format")

    def play_after(
            self,
            audio_file: str,
            event_name: str,
            stopped: callable,
            signals=None
    ):
        """
        Start audio playback using QtMultimedia

        :param audio_file: Path to audio file
        :param event_name: Event name to emit on playback
        :param stopped: Callback to call when playback is stopped
        :param signals: Signals to emit on playback
        :return: True if started
        """
        from PySide6.QtMultimedia import QMediaPlayer, QAudioOutput, QMediaDevices
        from PySide6.QtCore import QUrl, QTimer
        if signals is not None:
            signals.playback.emit(event_name)

        self.audio_output = QAudioOutput()
        self.audio_output.setVolume(1.0)

        devices = QMediaDevices.audioOutputs()
        if devices:
            try:
                num_device = int(self.window.core.config.get('audio.output.device', 0))
            except Exception:
                num_device = 0
            selected_device = devices[num_device] if num_device < len(devices) else devices[0]
            self.audio_output.setDevice(selected_device)

        if self.AUTO_CONVERT_TO_WAV:
            if audio_file.lower().endswith('.mp3'):
                tmp_dir = self.window.core.audio.get_cache_dir()
                base_name = os.path.splitext(os.path.basename(audio_file))[0]
                dst_file = os.path.join(tmp_dir, "_" + base_name + ".wav")
                wav_file = self.window.core.audio.mp3_to_wav(audio_file, dst_file)
                if wav_file:
                    audio_file = wav_file

        def check_stop():
            if stopped():
                self.player.stop()
                self.stop_timers()
                signals.volume_changed.emit(0)
            else:
                if self.player:
                    if self.player.playbackState() == QMediaPlayer.StoppedState:
                        self.player.stop()
                        self.stop_timers()
                        signals.volume_changed.emit(0)

        self.envelope = self.calculate_envelope(audio_file, self.chunk_ms)
        self.player = QMediaPlayer()
        self.player.setAudioOutput(self.audio_output)
        self.player.setSource(QUrl.fromLocalFile(audio_file))
        self.player.play()

        self.playback_timer = QTimer()
        self.playback_timer.setInterval(100)
        self.playback_timer.timeout.connect(check_stop)
        self.volume_timer = QTimer(self)
        self.volume_timer.setInterval(10)  # every 100 ms
        self.volume_timer.timeout.connect(
            lambda: self.update_volume(signals)
        )

        self.playback_timer.start()
        self.volume_timer.start()
        signals.volume_changed.emit(0)

    def stop_timers(self):
        """
        Stop playback timers.
        """
        if self.playback_timer is not None:
            self.playback_timer.stop()
            self.playback_timer = None
        if self.volume_timer is not None:
            self.volume_timer.stop()
            self.volume_timer = None

    def play(
            self,
            audio_file: str,
            event_name: str,
            stopped: callable,
            signals=None
    ):
        """
        Start audio playback.

        :param audio_file: Path to audio file
        :param event_name: Event name to emit on playback
        :param stopped: Callback to call when playback is stopped
        :param signals: Signals object to emit playback events.
        :return: True if started successfully.
        """
        self.play_after(audio_file, event_name, stopped, signals)

    def stop_playback(self, signals=None) -> bool:
        """
        Stop audio playback.

        :param signals: Signals object to emit stop event.
        :return: True if stopped successfully.
        """
        if self.player is not None:
            self.player.stop()
            if self.playback_timer is not None:
                self.playback_timer.stop()
        return False

    def calculate_envelope(
            self,
            audio_file: str,
            chunk_ms: int = 100
    ) -> list:
        """
        Calculate the volume envelope of an audio file.

        :param audio_file: Path to the audio file
        :param chunk_ms: Size of each chunk in milliseconds
        """
        import numpy as np
        audio = AudioSegment.from_file(audio_file)
        max_amplitude = 32767
        envelope = []

        for ms in range(0, len(audio), chunk_ms):
            chunk = audio[ms:ms + chunk_ms]
            rms = chunk.rms
            if rms > 0:
                db = 20 * np.log10(rms / max_amplitude)
            else:
                db = -60
            db = max(-60, min(0, db))
            volume = ((db + 60) / 60) * 100
            envelope.append(volume)

        return envelope

    def update_volume(self, signals=None):
        """
        Update the volume based on the current position in the audio file.

        :param signals: Signals object to emit volume changed event.
        """
        pos = self.player.position()
        index = int(pos / self.chunk_ms)
        if index < len(self.envelope):
            volume = self.envelope[index]
        else:
            volume = 0
        signals.volume_changed.emit(volume)

    def get_input_devices(self) -> List[Tuple[int, str]]:
        """
        Get input devices

        :return devices list: [(id, name)]
        """
        from bs4 import UnicodeDammit
        from PySide6.QtMultimedia import QMediaDevices
        devices = QMediaDevices.audioInputs()
        devices_list = []
        for index, device in enumerate(devices):
            dammit = UnicodeDammit(device.description())
            devices_list.append((index, dammit.unicode_markup))
        return devices_list

    def get_output_devices(self) -> List[Tuple[int, str]]:
        """
        Get output devices

        :return devices list: [(id, name)]
        """
        from bs4 import UnicodeDammit
        from PySide6.QtMultimedia import QMediaDevices
        devices = QMediaDevices.audioOutputs()
        devices_list = []
        for index, device in enumerate(devices):
            dammit = UnicodeDammit(device.description())
            devices_list.append((index, dammit.unicode_markup))
        return devices_list

    def get_default_input_device(self) -> tuple:
        """
        Retrieve the default input device using PyAudio.

        :return: (device_id, error_message) tuple
        """
        from PySide6.QtMultimedia import QMediaDevices
        default_device = QMediaDevices.defaultAudioInput()
        devices = QMediaDevices.audioInputs()
        try:
            index = devices.index(default_device)
        except ValueError:
            index = None
        return index, None

    def get_default_output_device(self):
        """
        Get the default audio output device ID using PySide6's QMediaDevices.

        :return: (device_id, None) if successful, (None, error_message) if failed
        """
        from PySide6.QtMultimedia import QMediaDevices
        default_device = QMediaDevices.defaultAudioOutput()
        devices = QMediaDevices.audioOutputs()
        try:
            index = devices.index(default_device)
        except ValueError:
            index = None
        return index, None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.19 07:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QObject, Signal, Slot, QEvent, QTimer
from PySide6.QtWebChannel import QWebChannel
from PySide6.QtWebEngineCore import QWebEngineSettings, QWebEnginePage, QWebEngineProfile
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu

from pygpt_net.core.events import RenderEvent
from pygpt_net.item.ctx import CtxMeta
from pygpt_net.core.text.web_finder import WebFinder
from pygpt_net.ui.widget.tabs.layout import FocusEventFilter
from pygpt_net.utils import trans, mem_clean

SHARED_PROFILE = None

import pygpt_net.icons_rc

class ChatWebOutput(QWebEngineView):
    def __init__(self, window=None):
        """
        HTML output (WebEngine)

        :param window: Window instance
        """
        super(ChatWebOutput, self).__init__(window)
        self.window = window
        self.finder = WebFinder(window, self)
        self.loadFinished.connect(self.on_page_loaded)
        self.customContextMenuRequested.connect(self.on_context_menu)
        self.signals = WebEngineSignals(self)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.filter = FocusEventFilter(self, self.on_focus)
        self.installEventFilter(self)
        self.plain = None
        self.html_content = None
        self.meta = None
        self.tab = None
        self.setProperty('class', 'layout-output-web')

        # OpenGL widgets
        self._glwidget = None
        self._glwidget_filter_installed = False

        # set the page with a shared profile
        self.setUpdatesEnabled(False)  # disable updates until the page is set, re-enable in `on_page_loaded`
        self.setPage(CustomWebEnginePage(self.window, self, profile=self._make_shared_profile()))

    def _make_shared_profile(self) -> QWebEngineProfile:
        """
        Create a shared QWebEngineProfile

        :return: QWebEngineProfile - shared profile instance
        """
        global SHARED_PROFILE
        if not SHARED_PROFILE:
            prof = QWebEngineProfile("app", None)
            # prof.setHttpCacheType(QWebEngineProfile.MemoryHttpCache)
            # prof.setHttpCacheMaximumSize(32 * 1024 * 1024)  # 32MB
            prof.setPersistentCookiesPolicy(QWebEngineProfile.NoPersistentCookies)
            prof.setSpellCheckEnabled(False)
            SHARED_PROFILE = prof
        return SHARED_PROFILE

    def _detach_gl_event_filter(self):
        """Detach OpenGL widget event filter if installed"""
        if self._glwidget and self._glwidget_filter_installed:
            try:
                self._glwidget.removeEventFilter(self)
            except Exception:
                pass
        self._glwidget = None
        self._glwidget_filter_installed = False

    def _release_profile_after_page(self, page, profile):
        """
        Release profile after page is destroyed

        :param page: QWebEnginePage - page to check
        :param profile: QWebEngineProfile - profile to release
        """
        if not profile or profile is QWebEngineProfile.defaultProfile():
            return
        if page:
            try:
                page.destroyed.connect(profile.deleteLater)
                return
            except Exception:
                pass
        try:
            profile.deleteLater()
        except Exception:
            pass

    def _teardown_page(self, page: QWebEnginePage):
        """
        Teardown page to clean up resources

        :param page: QWebEnginePage - page to teardown
        """
        if not page:
            return
        try:
            # detach the channel from the page to break JS<->Python references
            page.cleanup()
            page.setWebChannel(None)
        except Exception as e:
            pass

        # bridge, channel, and signals have parent=page, so deleteLater of the page will clean them up
        page.deleteLater()

    def _create_profile(self, parent=None) -> QWebEngineProfile:
        """
        Create a new QWebEngineProfile with off-the-record settings

        :param parent: QWidget - parent widget
        :return: QWebEngineProfile - new profile instance
        """
        p = QWebEngineProfile(parent or self)
        try:
            p.setHttpCacheMaximumSize(0)
            p.setPersistentCookiesPolicy(QWebEngineProfile.NoPersistentCookies)
            p.setSpellCheckEnabled(False)
        except Exception:
            pass
        return p

    def hard_reset(self, focus_after=True):
        """
        Hard reset of the view, creating a new instance of ChatWebOutput

        :param focus_after: bool - whether to focus the new view after reset
        """
        parent = self.parentWidget()
        if parent is None or parent.layout() is None:
            new_view = ChatWebOutput(self.window)
            new_view.set_tab(self.tab)
            new_view.set_meta(self.meta)
            QTimer.singleShot(0, self.on_delete)
            return new_view

        layout = parent.layout()
        idx = layout.indexOf(self)
        if idx < 0:
            idx = layout.count()

        new_view = ChatWebOutput(self.window)
        new_view.setProperty('class', self.property('class') or 'layout-output-web')
        new_view.set_tab(self.tab)
        new_view.set_meta(self.meta)

        try:
            z = self.get_zoom_value()
            p = new_view.page()
            if p:
                p.setZoomFactor(z)
        except Exception:
            pass

        layout.insertWidget(idx, new_view)
        layout.removeWidget(self)
        self.hide()

        QTimer.singleShot(0, self.on_delete)
        if focus_after:
            try:
                new_view.setFocus()
            except Exception:
                pass
        try:
            mem_clean()
        except Exception:
            pass

        return new_view

    def resetPage(self):
        """Reset current page without creating a new one"""
        self.plain = None
        self.html_content = None
        p = self.page()
        if not p:
            self.setPage(CustomWebEnginePage(self.window, self, profile=SHARED_PROFILE))
            p = self.page()

        p.runJavaScript(
            f"""clean();"""
        )
        try:
            p.history().clear()
        except Exception:
            pass
        mem_clean()

    def on_delete(self):
        """Clean up on delete"""
        self._detach_gl_event_filter()
        if self.finder:
            try:
                self.finder.disconnect()
            except Exception:
                pass
            self.finder = None

        self.tab = None
        self.meta = None

        # remove the page and profile
        page = self.page()
        prof = getattr(self, "_profile", None)
        if page:
            self._teardown_page(page)

        # safely unhook signals (may not have been hooked)
        for sig, slot in (
            (self.loadFinished, self.on_page_loaded),
            (self.customContextMenuRequested, self.on_context_menu),
            (self.signals.save_as, getattr(self.window.controller.chat.render, "handle_save_as", None)),
            (self.signals.audio_read, getattr(self.window.controller.chat.render, "handle_audio_read", None)),
        ):
            if slot:
                try:
                    sig.disconnect(slot)
                except Exception:
                    pass

        self.deleteLater()

    def eventFilter(self, source, event):
        """
        Event filter to handle child added events and mouse button presses

        :param source: QWidget - source of the event
        :param event: QEvent - event to filter
        """
        if event.type() == QEvent.ChildAdded and source is self and event.child().isWidgetType():
            self._detach_gl_event_filter()
            self._glwidget = event.child()
            try:
                self._glwidget.installEventFilter(self)
                self._glwidget_filter_installed = True
            except Exception:
                self._glwidget = None
                self._glwidget_filter_installed = False

        elif event.type() == QEvent.Type.MouseButtonPress:
            try:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
            except Exception:
                pass

        return super().eventFilter(source, event)

    def on_focus(self, widget):
        """
        On widget clicked

        :param widget: QWidget - widget that received focus
        """
        if self.tab is not None:
            self.window.controller.ui.tabs.on_column_focus(self.tab.column_idx)
        self.setFocus()

    def set_tab(self, tab):
        """
        Set tab for this output

        :param tab: Tab instance
        """
        self.tab = tab

    def set_meta(self, meta: CtxMeta):
        """
        Set meta for this output

        :param meta: CtxMeta instance
        """
        self.meta = meta

    def set_plaintext(self, text: str):
        """
        Set plain text content

        :param text: str - plain text content
        """
        self.plain = text

    def set_html_content(self, html: str):
        """
        Set HTML content

        :param html: str - HTML content
        """
        self.html_content = "<html>" + html + "</html>"

    def on_context_menu(self, position):
        """
        Context menu event

        :param position: QPoint - position of the context menu
        """
        menu = QMenu(self)

        has_selection = self.page().hasSelection()

        if has_selection:
            # copy
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.copy'), self)
            action.triggered.connect(self.copy_selected_text)
            menu.addAction(action)

            # audio read (get text only on click, don't copy immediately)
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self._read_selected_text)
            menu.addAction(action)

            # copy to
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(self, self.get_selected_text())
            menu.addMenu(copy_to_menu)

            # save as (selected) - get selection at the moment of click
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(self._save_selected_txt)
            menu.addAction(action)
        else:
            # select all
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.select_all'), self)
            action.triggered.connect(self.select_all_text)
            menu.addAction(action)

            # save as (all) - plain (lazy normalization only on click)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (text)", self)
            action.triggered.connect(self._save_as_text)
            menu.addAction(action)

            # save as (all) - html
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (html)", self)
            action.triggered.connect(self._save_as_html)
            menu.addAction(action)

        action = QAction(QIcon(":/icons/search.svg"), trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        menu.addAction(action)

        menu.exec_(self.mapToGlobal(position))

    @Slot()
    def _save_selected_txt(self):
        """Save selected content as text file"""
        self.signals.save_as.emit(self.get_selected_text(), 'txt')

    @Slot()
    def _read_selected_text(self):
        """
        Read selected text using text-to-speech
        """
        selected_text = self.get_selected_text()
        if selected_text:
            self.signals.audio_read.emit(selected_text)

    @Slot()
    def _save_as_text(self):
        """
        Save current content as text file
        """
        # TODO: normalize text (remove extra spaces, newlines, etc.)
        self.page().toPlainText(lambda txt: self.signals.save_as.emit(txt, 'txt'))

    @Slot()
    def _save_as_html(self):
        """
        Save current content as HTML file
        """
        self.page().toHtml(lambda html: self.signals.save_as.emit(html, 'html'))

    def update_zoom(self):
        """Update zoom from config"""
        try:
            if self.window.core.config.has("zoom"):
                p = self.page()
                if p:
                    p.setZoomFactor(self.window.core.config.get("zoom"))
        except Exception:
            pass

    def on_focus_js(self):
        """Focus JavaScript"""
        if self.tab is not None:
            self.window.controller.ui.tabs.on_column_focus(self.tab.column_idx)

    def get_zoom_value(self) -> float:
        """
        Get current zoom value

        :return: zoom factor
        """
        p = self.page()
        return p.zoomFactor() if p else 1.0

    def reset_current_content(self):
        """Reset current content"""
        self.plain = ""
        self.html_content = ""

    def on_page_loaded(self, success):
        """
        Page loaded event handler

        :param success: bool - True if page loaded successfully, False otherwise
        """
        if success:
            event = RenderEvent(RenderEvent.ON_PAGE_LOAD, {
                "meta": self.meta,
                "tab": self.tab,
            })
            self.window.dispatch(event)
            self.setUpdatesEnabled(True)

    def get_selected_text(self) -> str:
        p = self.page()
        return p.selectedText() if p else ""

    def copy_selected_text(self):
        p = self.page()
        if p:
            p.triggerAction(QWebEnginePage.Copy)

    def select_all_text(self):
        p = self.page()
        if p:
            p.triggerAction(QWebEnginePage.SelectAll)

    def unselect_text(self):
        p = self.page()
        if p:
            p.triggerAction(QWebEnginePage.Unselect)

    def find_open(self):
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        if self.finder:
            self.finder.clear()

    def focusInEvent(self, e):
        super(ChatWebOutput, self).focusInEvent(e)
        self.window.controller.finder.focus_in(self.finder)


class CustomWebEnginePage(QWebEnginePage):
    """Custom WebEnginePage to handle web events"""
    def __init__(self, window, view, profile: QWebEngineProfile = None):

        # use the profile if provided, otherwise the default
        if profile is not None:
            super(CustomWebEnginePage, self).__init__(profile, view)
        else:
            super(CustomWebEnginePage, self).__init__(view)

        self.window = window
        self.view = view

        # signals have parent=page (automatic cleanup)
        self.signals = WebEnginePageSignals(self)

        self.findTextFinished.connect(self.on_find_finished)
        self.zoomFactorChanged.connect(self.on_view_changed)
        self.selectionChanged.connect(self.on_selection_changed)

        s = self.settings()
        s.setAttribute(QWebEngineSettings.LocalContentCanAccessFileUrls, True)
        s.setAttribute(QWebEngineSettings.LocalContentCanAccessRemoteUrls, True)
        s.setFontFamily(QWebEngineSettings.StandardFont, 'Lato')
        s.setFontFamily(QWebEngineSettings.FixedFont, 'Monaspace Neon')
        s.setFontFamily(QWebEngineSettings.SerifFont, 'Monaspace Neon')

        if self.window.core.config.has("zoom"):
            self.setZoomFactor(self.window.core.config.get("zoom"))

        # bridge Python <> JavaScript
        # KEY: parent=page, -> disappears along with the page
        self.bridge = Bridge(self.window, parent=self)
        self.channel = QWebChannel(self)
        self.channel.registerObject("bridge", self.bridge)
        self.setWebChannel(self.channel)

    def on_find_finished(self, result):
        current = int(result.activeMatch())
        num = int(result.numberOfMatches())
        self.view.finder.current_match_index = current
        self.view.finder.matches = num
        self.view.finder.on_find_finished()

    def on_view_changed(self):
        zoom = self.zoomFactor()
        self.window.core.config.set("zoom", zoom)
        option = self.window.controller.settings.editor.get_option('zoom')
        option['value'] = zoom
        self.window.controller.config.apply(
            parent_id='config',
            key='zoom',
            option=option,
        )

    def on_selection_changed(self):
        pass

    def acceptNavigationRequest(self, url, _type, isMainFrame):
        if _type == QWebEnginePage.NavigationTypeLinkClicked:
            self.window.core.filesystem.url.handle(url)
            return False
        return super().acceptNavigationRequest(url, _type, isMainFrame)

    def javaScriptConsoleMessage(self, level, message, line_number, source_id):
        # print("[JS CONSOLE] Line", line_number, ":", message)
        self.signals.js_message.emit(line_number, message, source_id)  # handled in debug controller

    def cleanup(self):
        """Cleanup method to release resources"""
        try:
            self.findTextFinished.disconnect()
            self.zoomFactorChanged.disconnect()
            self.selectionChanged.disconnect()
        except Exception:
            pass

        if self.bridge:
            try:
                self.bridge.cleanup()
            except Exception:
                pass
            self.bridge = None

        if self.channel:
            try:
                self.channel.unregisterObject("bridge")
            except Exception:
                pass
            self.channel = None

        if self.signals:
            try:
                self.signals.deleteLater()
            except Exception:
                pass
            self.signals = None

        # delete the page object
        self.deleteLater()


class Bridge(QObject):
    """Bridge between Python and JavaScript"""
    def __init__(self, window, parent=None):
        super(Bridge, self).__init__(parent)
        self.window = window

    chunk = Signal(str, str, str, bool, bool)  # name, buffer, chunk, replace, is_code
    readyChanged = Signal(bool)

    @Slot()
    def js_ready(self):
        self.readyChanged.emit(True)

    @Slot(str)
    def log(self, text: str):
        print(f"JS log: {text}")

    @Slot(str)
    def copy_text(self, text: str):
        self.window.controller.ctx.extra.copy_code_text(text)

    @Slot(str)
    def preview_text(self, text: str):
        self.window.controller.ctx.extra.preview_code_text(text)

    @Slot(str)
    def run_text(self, text: str):
        self.window.controller.ctx.extra.run_code_text(text)

    @Slot(int)
    def update_scroll_position(self, pos: int):
        self.window.controller.chat.render.scroll = pos

    def cleanup(self):
        """Cleanup method to release resources"""
        if self.window:
            try:
                self.window = None
            except Exception:
                pass

        # delete the bridge object
        self.deleteLater()


class WebEngineSignals(QObject):
    save_as = Signal(str, str)
    audio_read = Signal(str)


class WebEnginePageSignals(QObject):
    js_message = Signal(int, str, str)  # on JavaScript message


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.19 07:00:00                  #
# ================================================== #

import io
from pygpt_net.utils import trans


class PidData:

    def __init__(self, pid, meta=None):
        """Pid Data"""
        self.pid = pid
        self.meta = meta
        self.images_appended = []
        self.urls_appended = []
        self.files_appended = []
        self._buffer = io.StringIO()
        self._live_buffer = io.StringIO()
        self._html = io.StringIO()
        self._document = io.StringIO()
        self.is_cmd = False
        self.initialized = False
        self.loaded = False
        self.item = None
        self.use_buffer = False
        self.name_user = trans("chat.name.user")
        self.name_bot = trans("chat.name.bot")
        self.last_time_called = 0
        self.cooldown = 1 / 6
        self.throttling_min_chars = 5000
        self.header = None

    @property
    def buffer(self) -> str:
        return self._buffer.getvalue()

    @buffer.setter
    def buffer(self, value: str):
        if value is None or value == "":
            self._buffer.close()
            self._buffer = io.StringIO()
        self._buffer.seek(0)
        self._buffer.truncate(0)
        if value:
            self._buffer.write(value)

    def append_buffer(self, text: str):
        self._buffer.write(text)

    @property
    def live_buffer(self) -> str:
        return self._live_buffer.getvalue()

    @live_buffer.setter
    def live_buffer(self, value: str):
        if value is None or value == "":
            self._live_buffer.close()
            self._live_buffer = io.StringIO()
        self._live_buffer.seek(0)
        self._live_buffer.truncate(0)
        if value:
            self._live_buffer.write(value)

    def append_live_buffer(self, text: str):
        self._live_buffer.write(text)

    @property
    def html(self) -> str:
        return self._html.getvalue()

    @html.setter
    def html(self, value: str):
        if value is None or value == "":
            self._html.close()
            self._html = io.StringIO()
        self._html.seek(0)
        self._html.truncate(0)
        if value:
            self._html.write(value)

    def append_html(self, text: str):
        self._html.write(text)

    @property
    def document(self) -> str:
        return self._document.getvalue()

    @document.setter
    def document(self, value: str):
        if value is None or value == "":
            self._document.close()
            self._document = io.StringIO()
        self._document.seek(0)
        self._document.truncate(0)
        if value:
            self._document.write(value)

    def append_document(self, text: str):
        self._document.write(text)

    def clear(self, all: bool = False):
        """
        Clear buffers and other data

        :param all: If True, clear all data, otherwise only buffers
        """
        for name in ("_buffer", "_live_buffer", "_html", "_document"):
            try:
                getattr(self, name).close()
            except Exception:
                pass
            setattr(self, name, io.StringIO())

        if all:
            self.item = None
            self.images_appended.clear()
            self.urls_appended.clear()
            self.files_appended.clear()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.26 18:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QComboBox

from pygpt_net.ui.widget.element.labels import TitleLabel
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ComputerEnv:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'computer_env'

    def setup(self) -> QWidget:
        """
        Setup indexes

        :return: QWidget
        """
        layout = self.setup_env()
        self.window.ui.nodes['env.widget'] = QWidget()
        self.window.ui.nodes['env.widget'].setLayout(layout)
        return self.window.ui.nodes['env.widget']

    def setup_env(self) -> QVBoxLayout:
        """
        Setup list of environments

        :return: QVBoxLayout
        """

        # label
        self.window.ui.nodes['env.label'] = TitleLabel(trans("toolbox.env.label"))

        # list
        self.window.ui.nodes[self.id] = QComboBox()
        self.window.ui.nodes[self.id].addItem("Browser", "browser")
        self.window.ui.nodes[self.id].addItem("Windows", "windows")
        self.window.ui.nodes[self.id].addItem("Linux", "linux")
        self.window.ui.nodes[self.id].addItem("Mac", "mac")
        self.window.ui.nodes[self.id].setMinimumWidth(40)

        # on change signal
        self.window.ui.nodes[self.id].currentIndexChanged.connect(
            lambda index: self.window.controller.ui.on_computer_env_changed(
                self.window.ui.nodes[self.id].itemData(index)))

        # rows
        layout = QHBoxLayout()
        layout.addWidget(self.window.ui.nodes['env.label'], 0)  # label
        layout.addWidget(self.window.ui.nodes[self.id], 1)  # combobox
        layout.setContentsMargins(2, 5, 5, 5)
        return layout


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.12 19:00:00                  #
# ================================================== #

import io

class PidData:

    def __init__(self, pid, meta=None):
        """Pid Data"""
        self.pid = pid
        self.meta = meta
        self.images_appended = []
        self.urls_appended = []
        self.files_appended = []
        self._buffer = io.StringIO()
        self.is_cmd = False

    @property
    def buffer(self) -> str:
        return self._buffer.getvalue()

    @buffer.setter
    def buffer(self, value: str):
        self._buffer.seek(0)
        self._buffer.truncate(0)
        if value:
            self._buffer.write(value)

    def append_buffer(self, text: str):
        self._buffer.write(text)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.16 00:00:00                  #
# ================================================== #

import datetime
import functools

from PySide6 import QtWidgets, QtCore, QtGui
from PySide6.QtCore import Qt, QPoint, QItemSelectionModel
from PySide6.QtGui import QAction, QIcon, QColor, QPixmap, QStandardItem
from PySide6.QtWidgets import QMenu
from overrides import overrides

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ContextList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Context select menu

        :param window: main window
        :param id: input id
        """
        super(ContextList, self).__init__(window)
        self.window = window
        self.id = id
        self.expanded_items = set()
        self._icons = {
            'add': QIcon(":/icons/add.svg"),
            'edit': QIcon(":/icons/edit.svg"),
            'delete': QIcon(":/icons/delete.svg"),
            'chat': QIcon(":/icons/chat.svg"),
            'copy': QIcon(":/icons/copy.svg"),
            'close': QIcon(":/icons/close.svg"),
            'pin': QIcon(":/icons/pin.svg"),
            'clock': QIcon(":/icons/clock.svg"),
            'db': QIcon(":/icons/db.svg"),
            'folder': QIcon(":/icons/folder_filled.svg"),
            'attachment': QIcon(":/icons/attachment.svg"),
        }
        self._color_icon_cache = {}
        self.setItemDelegate(ImportantItemDelegate(self, self._icons['attachment']))
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)
        self._backup_selection = None
        self.restore_after_ctx_menu = True

    @property
    def _model(self):
        return self.window.ui.models['ctx.list']

    @property
    def _view(self):
        return self.window.ui.nodes['ctx.list']

    def _color_icon(self, color: QColor) -> QIcon:
        key = color.rgba()
        icon = self._color_icon_cache.get(key)
        if icon is None:
            pixmap = QPixmap(16, 16)
            pixmap.fill(color)
            icon = QIcon(pixmap)
            self._color_icon_cache[key] = icon
        return icon

    def click(self, index):
        """
        Click event (override, connected in BaseList class)

        :param index: index
        """
        item = self._model.itemFromIndex(index)
        if item is None or not hasattr(item, 'isFolder'):
            return
        if item.isFolder:
            self.window.controller.ctx.set_group(item.id)
            if self._view.isExpanded(index):
                self.expanded_items.discard(item.id)
                self._view.collapse(index)
            else:
                self.expanded_items.add(item.id)
                self._view.expand(index)
        else:
            self.window.controller.ctx.select_by_id(item.id)

    def expand_group(self, id):
        """
        Expand group

        :param id: group id
        """
        for i in range(self._model.rowCount()):
            item = self._model.item(i)
            if isinstance(item, GroupItem) and item.id == id:
                index = self._model.indexFromItem(item)
                self._view.expand(index)
                self.expanded_items.add(id)
                break

    def dblclick(self, index):
        """
        Double click event

        :param index: index
        """
        print("dblclick")

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            if not index.isValid():
                self.window.controller.ctx.unselect()
                return
            super().mousePressEvent(event)
        elif event.button() == Qt.RightButton:
            index = self.indexAt(event.pos())
            if index.isValid():
                sel = self.selectionModel()
                self._backup_selection = list(sel.selectedIndexes())
                sel.clearSelection()
                sel.select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
            event.accept()
        else:
            super().mousePressEvent(event)

    def show_context_menu(self, pos: QPoint):
        """
        Context menu event

        :param pos: QPoint
        """
        global_pos = self.viewport().mapToGlobal(pos)
        index = self.indexAt(pos)
        item = self._model.itemFromIndex(index)

        if item is not None and index.isValid() and hasattr(item, 'id'):
            idx = item.row()
            id_value = item.id

            if hasattr(item, 'isFolder') and item.isFolder:
                menu = QMenu(self)
                a_new = menu.addAction(self._icons['add'], trans('action.ctx.new'))
                a_new.triggered.connect(functools.partial(self.window.controller.ctx.new_in_group, force=False, group_id=id_value))
                a_rename = menu.addAction(self._icons['edit'], trans('action.rename'))
                a_rename.triggered.connect(functools.partial(self.window.controller.ctx.rename_group, id_value))
                a_delete = menu.addAction(self._icons['delete'], trans('action.group.delete.only'))
                a_delete.triggered.connect(functools.partial(self.window.controller.ctx.delete_group, id_value))
                a_delete_all = menu.addAction(self._icons['delete'], trans('action.group.delete.all'))
                a_delete_all.triggered.connect(functools.partial(self.window.controller.ctx.delete_group_all, id_value))
                if idx >= 0:
                    menu.exec(global_pos)
            else:
                ctx_id = id_value
                ctx = self.window.core.ctx.get_meta_by_id(ctx_id)
                if ctx is None:
                    return

                is_important = ctx.important

                menu = QMenu(self)
                a_open = menu.addAction(self._icons['chat'], trans('action.open'))
                a_open.triggered.connect(functools.partial(self.action_open, ctx_id, idx))

                a_open_new_tab = menu.addAction(self._icons['chat'], trans('action.open_new_tab'))
                a_open_new_tab.triggered.connect(functools.partial(self.action_open_new_tab, ctx_id, idx))

                a_rename = menu.addAction(self._icons['edit'], trans('action.rename'))
                a_rename.triggered.connect(functools.partial(self.action_rename, ctx_id))

                a_duplicate = menu.addAction(self._icons['copy'], trans('action.duplicate'))
                a_duplicate.triggered.connect(functools.partial(self.action_duplicate, ctx_id))

                if is_important:
                    a_pin = menu.addAction(self._icons['pin'], trans('action.unpin'))
                    a_pin.triggered.connect(functools.partial(self.action_unpin, ctx_id))
                else:
                    a_pin = menu.addAction(self._icons['pin'], trans('action.pin'))
                    a_pin.triggered.connect(functools.partial(self.action_pin, ctx_id))

                a_delete = menu.addAction(self._icons['delete'], trans('action.delete'))
                a_delete.triggered.connect(functools.partial(self.action_delete, ctx_id))

                colors = self.window.controller.ui.get_colors()
                set_label_menu = menu.addMenu(trans('calendar.day.label'))
                for status_id, status_info in colors.items():
                    name = trans('calendar.day.' + status_info['label']) if status_id != 0 else '-'
                    icon = self._color_icon(status_info['color'])
                    status_action = set_label_menu.addAction(icon, name)
                    status_action.triggered.connect(
                        functools.partial(self.action_set_label, ctx_id, status_id)
                    )

                idx_menu = QMenu(trans('action.idx'), self)
                idxs = self.window.core.config.get('llama.idx.list')
                store = self.window.core.idx.get_current_store()
                if idxs:
                    for idx_dict in idxs:
                        index_id = idx_dict['id']
                        name = idx_dict['name'] + " (" + idx_dict['id'] + ")"
                        action = idx_menu.addAction(self._icons['db'], "IDX: " + name)
                        action.triggered.connect(functools.partial(self.action_idx, ctx_id, index_id))

                    if ctx.indexed is not None and ctx.indexed > 0:
                        if store in ctx.indexes:
                            store_indexes = ctx.indexes[store]
                            idx_menu.addSeparator()
                            for store_index in store_indexes:
                                action = idx_menu.addAction(self._icons['delete'], trans("action.idx.remove") + ": " + store_index)
                                action.triggered.connect(
                                    functools.partial(self.action_idx_remove, store_index, ctx_id)
                                )
                    menu.addMenu(idx_menu)

                group_menu = QMenu(trans('action.move_to'), self)
                groups = self.window.core.ctx.get_groups()

                action = group_menu.addAction(self._icons['add'], trans("action.group.new"))
                action.triggered.connect(functools.partial(self.window.controller.ctx.new_group, ctx_id))

                if groups:
                    group_menu.addSeparator()

                for group_id, group in groups.items():
                    action = group_menu.addAction(self._icons['folder'], group.name)
                    action.triggered.connect(functools.partial(self.window.controller.ctx.move_to_group, ctx_id, group_id))

                if groups:
                    group_menu.addSeparator()

                if ctx.group_id is not None and ctx.group_id > 0:
                    group_name = str(ctx.group_id)
                    if ctx.group_id in groups:
                        group_name = groups[ctx.group_id].name
                    action = group_menu.addAction(self._icons['delete'], trans("action.group.remove") + ": " + group_name)
                    action.triggered.connect(functools.partial(self.window.controller.ctx.remove_from_group, ctx_id))

                menu.addMenu(group_menu)

                a_copy_id = menu.addAction(self._icons['copy'], trans('action.ctx_copy_id') + " @" + str(ctx_id))
                a_copy_id.triggered.connect(functools.partial(self.action_copy_id, ctx_id))

                if ctx.indexed is not None and ctx.indexed > 0:
                    suffix = ""
                    if ctx.updated > ctx.indexed:
                        suffix = " *"
                    dt = datetime.datetime.fromtimestamp(ctx.indexed).strftime("%Y-%m-%d %H:%M")
                    action = menu.addAction(self._icons['clock'], trans('action.ctx.indexed') + ": " + dt + suffix)
                    action.setEnabled(False)

                a_reset = menu.addAction(self._icons['close'], trans('action.ctx_reset'))
                a_reset.triggered.connect(functools.partial(self.action_reset, ctx_id))

                if idx >= 0:
                    self.window.controller.ctx.set_selected(ctx_id)
                    menu.exec(global_pos)

        self.store_scroll_position()

        if self.restore_after_ctx_menu:
            if self._backup_selection is not None:
                sel = self.selectionModel()
                sel.clearSelection()
                for sel_idx in self._backup_selection:
                    sel.select(sel_idx, QItemSelectionModel.Select | QItemSelectionModel.Rows)
                self._backup_selection = None

        self.restore_after_ctx_menu = True
        self.restore_scroll_position()

    def action_open(self, id: int, idx: int = None):
        """
        Open context action handler

        :param id: context id
        :param idx: index id (optional)
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.load(id, select_idx=idx)

    def action_open_new_tab(self, id: int, idx: int = None):
        """
        Open context action handler in nowej karcie

        :param id: context id
        :param idx: index id (optional)
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.load(id, select_idx=idx, new_tab=True)

    def action_idx(self, id: int, idx: int):
        """
        Index with llama context action handler

        :param id: context id
        :param idx: index name
        """
        self.restore_after_ctx_menu = False
        self.window.controller.idx.indexer.index_ctx_meta(id, idx)

    def action_idx_remove(self, idx: str, meta_id: int):
        """
        Remove from index action handler

        :param idx: index id
        :param meta_id: meta id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.idx.indexer.index_ctx_meta_remove(idx, meta_id)

    def action_rename(self, id):
        """
        Rename action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.rename(id)

    def action_pin(self, id):
        """
        Pin action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.set_important(id, True)

    def action_unpin(self, id):
        """
        Unpin action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.set_important(id, False)

    def action_important(self, id):
        """
        Set as important action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.set_important(id)

    def action_duplicate(self, id):
        """
        Duplicate handler

        :param id: context id
        """
        self.window.controller.ctx.common.duplicate(id)

    def action_delete(self, id):
        """
        Delete action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.delete(id)

    def action_copy_id(self, id):
        """
        Copy ID tag action handler

        :param id: context id
        """
        self.window.controller.ctx.common.copy_id(id)

    def action_reset(self, id):
        """
        Reset action handler

        :param id: context id
        """
        self.restore_after_ctx_menu = False
        self.window.controller.ctx.common.reset(id)

    def action_set_label(self, id: int, label: int):
        """
        Set label action handler

        :param id: context id
        :param label: label id
        """
        self.window.controller.ctx.set_label(id, label)

    def selectionCommand(self, index, event=None):
        """
        Selection command
        :param index: Index
        :param event: Event
        """
        return super().selectionCommand(index, event)


class ImportantItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    Label color delegate
    """
    def __init__(self, parent=None, attachment_icon: QIcon = None):
        super().__init__(parent)
        self._attachment_icon = attachment_icon or QIcon(":/icons/attachment.svg")
        self._status_colors = {
            0: QColor(100, 100, 100),
            1: QColor(255, 0, 0),
            2: QColor(255, 165, 0),
            3: QColor(255, 255, 0),
            4: QColor(0, 255, 0),
            5: QColor(0, 0, 255),
            6: QColor(75, 0, 130),
            7: QColor(238, 130, 238),
        }
        self._pin_pen = QtGui.QPen(QtCore.Qt.black, 0.5, QtCore.Qt.SolidLine)

    def paint(self, painter, option, index):
        if index.parent().isValid():
            option.rect.adjust(15, 0, 0, 0)

        super(ImportantItemDelegate, self).paint(painter, option, index)

        data = index.data(QtCore.Qt.ItemDataRole.UserRole)
        if data:
            label = data.get("label", 0)
            is_important = data.get("is_important", False)
            is_attachment = data.get("is_attachment", False)

            painter.save()

            if is_attachment:
                icon_size = option.decorationSize or QtCore.QSize(16, 16)
                icon_pos = option.rect.right() - icon_size.width()
                y = option.rect.top() + (option.rect.height() - icon_size.height()) // 2
                icon_rect = QtCore.QRect(
                    icon_pos,
                    y,
                    icon_size.width(),
                    icon_size.height()
                )
                self._attachment_icon.paint(painter, icon_rect, QtCore.Qt.AlignCenter)

            if is_important:
                color = self.get_color_for_status(3)
                square_size = 3
                square_rect = QtCore.QRect(
                    option.rect.left(),
                    option.rect.top() + 2,
                    square_size,
                    square_size,
                )
                painter.setBrush(color)
                painter.setPen(self._pin_pen)
                painter.drawRect(square_rect)

            if label > 0:
                color = self.get_color_for_status(label)
                square_size = 5
                y = option.rect.center().y() - (square_size // 2) + 2
                square_rect = QtCore.QRect(
                    option.rect.left(),
                    y,
                    square_size,
                    square_size,
                )
                painter.setBrush(color)
                painter.setPen(QtCore.Qt.NoPen)
                painter.drawRect(square_rect)

            painter.restore()

    def get_color_for_status(self, status: int) -> QColor:
        return self._status_colors.get(status, self._status_colors[0])


class GroupItem(QStandardItem):
    def __init__(self, icon, name, id):
        super().__init__(icon, name)
        self.id = id
        self.name = name
        self.isFolder = True
        self.isPinned = False
        self.hasAttachments = False
        self.dt = None


class Item(QStandardItem):
    def __init__(self, name, id):
        super().__init__(name)
        self.id = id
        self.name = name
        self.isFolder = False
        self.isPinned = False
        self.dt = None


class SectionItem(QStandardItem):
    def __init__(self, title, group: bool = False):
        super().__init__(title)
        self.title = title
        self.group = group
        self.setSelectable(False)
        self.setEnabled(False)
        self.setTextAlignment(QtCore.Qt.AlignRight)
        font = self.font()
        font.setBold(True)
        self.setFont(font)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.19 07:00:00                  #
# ================================================== #

import os
from json import dumps as _json_dumps
from random import shuffle as _shuffle

from typing import Optional, List, Dict

from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .syntax_highlight import SyntaxHighlight

import pygpt_net.js_rc
import pygpt_net.css_rc
import pygpt_net.fonts_rc

class Body:

    NUM_TIPS = 13

    _HTML_P0 = """
            <!DOCTYPE html>
            <html>
            <head>
                <style>
                    """
    _HTML_P1 = """
                </style>
                <link rel="stylesheet" href="qrc:///css/katex.min.css">
                <script type="text/javascript" src="qrc:///qtwebchannel/qwebchannel.js"></script>
                <script type="text/javascript" src="qrc:///js/highlight.min.js"></script>
                <script type="text/javascript" src="qrc:///js/katex.min.js"></script>
                <script>
                const DEBUG_MODE = false;
                let scrollTimeout = null;
                let prevScroll = 0;
                let bridge;
                let streamQ = [];
                let streamRAF = 0;
                let pid = """
    _HTML_P2 = """                
                let collapsed_idx = [];
                let domOutputStream = document.getElementById('_append_output_');
                let domOutput = document.getElementById('_output_');
                let domInput = document.getElementById('_input_');
                let domLastCodeBlock = null;
                let domLastParagraphBlock = null;
                let htmlBuffer = "";
                let tips = """
    _HTML_P3 = """;
                let tips_hidden = false;

                let els = {};
                let highlightScheduled = false;
                let pendingHighlightRoot = null;
                let pendingHighlightMath = false;
                let scrollScheduled = false;

                // timers
                let tipsTimers = [];

                // clear previous references
                function resetEphemeralDomRefs() {
                    domLastCodeBlock = null;
                    domLastParagraphBlock = null;
                }
                function dropIfDetached() {
                    if (domLastCodeBlock && !domLastCodeBlock.isConnected) domLastCodeBlock = null;
                    if (domLastParagraphBlock && !domLastParagraphBlock.isConnected) domLastParagraphBlock = null;
                }
                function stopTipsTimers() {
                    tipsTimers.forEach(clearTimeout);
                    tipsTimers = [];
                }

                history.scrollRestoration = "manual";
                document.addEventListener('keydown', function(event) {
                    if (event.ctrlKey && event.key === 'f') {
                        window.location.href = 'bridge://open_find:' + pid;
                        event.preventDefault();
                    }
                    if (event.key === 'Escape') {
                        window.location.href = 'bridge://escape';
                        event.preventDefault();
                    }
                });
                document.addEventListener('click', function(event) {
                    if (event.target.tagName !== 'A' && !event.target.closest('a')) {
                        window.location.href = 'bridge://focus';
                    }
                });
                function log(text) {
                    if (bridge) {
                        bridge.log(text);
                    }
                }
                function initDomRefs() {
                    els.container = document.getElementById('container');
                    els.nodes = document.getElementById('_nodes_');
                    els.appendInput = document.getElementById('_append_input_');
                    els.appendOutputBefore = document.getElementById('_append_output_before_');
                    els.appendOutput = document.getElementById('_append_output_');
                    els.appendLive = document.getElementById('_append_live_');
                    els.footer = document.getElementById('_footer_');
                    els.loader = document.getElementById('_loader_');
                    els.tips = document.getElementById('tips');
                }
                function scheduleHighlight(root, withMath = true) {
                    const scope = root && root.nodeType === 1 ? root : document;
                    if (!pendingHighlightRoot || pendingHighlightRoot === document) {
                        pendingHighlightRoot = scope;
                    } else if (!pendingHighlightRoot.contains(scope)) {
                        pendingHighlightRoot = document;
                    }
                    if (withMath) pendingHighlightMath = true;
                    if (highlightScheduled) return;
                    highlightScheduled = true;
                    requestAnimationFrame(function() {
                        try {
                            highlightCodeInternal(pendingHighlightRoot || document, pendingHighlightMath);
                        } finally {
                            highlightScheduled = false;
                            pendingHighlightRoot = null;
                            pendingHighlightMath = false;
                        }
                    });
                }
                function highlightCodeInternal(root, withMath) {
                    (root || document).querySelectorAll('pre code:not(.hljs)').forEach(el => {
                        hljs.highlightElement(el);
                    });
                    if (withMath) {
                        renderMath(root);
                    }
                    if (DEBUG_MODE) log("execute highlight");
                    restoreCollapsedCode(root);
                }
                function highlightCode(withMath = true, root = null) {
                     if (DEBUG_MODE) log("queue highlight, withMath: " + withMath);
                    scheduleHighlight(root || document, withMath);
                }
                function hideTips() {
                    if (tips_hidden) return;
                    stopTipsTimers();
                    const t = els.tips || document.getElementById('tips');
                    if (t) t.style.display = 'none';
                    tips_hidden = true;
                }
                function showTips() {
                    if (tips_hidden) return;
                    if (tips.length === 0) return;
                    const t = els.tips || document.getElementById('tips');
                    if (t) t.style.display = 'block';
                    tips_hidden = false;
                }
                function cycleTips() {
                    if (tips_hidden) return;
                    if (tips.length === 0) return;
                    let currentTip = 0;
                    function showNextTip() {
                        if (tips_hidden) return;
                        const tipContainer = els.tips || document.getElementById('tips');
                        if (!tipContainer) return;
                        tipContainer.innerHTML = tips[currentTip];
                        tipContainer.classList.add('visible');
                        tipsTimers.push(setTimeout(function() {
                            if (tips_hidden) return;
                            tipContainer.classList.remove('visible');
                            tipsTimers.push(setTimeout(function(){
                                currentTip = (currentTip + 1) % tips.length;
                                showNextTip();
                            }, 1000));
                        }, 15000));
                    }
                    stopTipsTimers();
                    showNextTip();
                }
                function renderMath(root) {
                    if (DEBUG_MODE) log("execute math");
                    const scope = root || document;
                    const scripts = scope.querySelectorAll('script[type^="math/tex"]');
                    scripts.forEach(function(script) {
                        const displayMode = script.type.indexOf('mode=display') > -1;
                        const mathContent = script.textContent || script.innerText;
                        const element = document.createElement(displayMode ? 'div' : 'span');
                        try {
                          katex.render(mathContent, element, {
                            displayMode: displayMode,
                            throwOnError: false
                          });
                        } catch (err) {
                          element.textContent = mathContent;
                        }
                        const parent = script.parentNode;
                        if (parent) parent.replaceChild(element, script);
                      });
                }
                function isNearBottom(marginPx = 100) {
                    const el = document.scrollingElement || document.documentElement;
                    const distanceToBottom = el.scrollHeight - el.clientHeight - el.scrollTop;
                    return distanceToBottom <= marginPx;
                }
                function scheduleScroll(live = false) {
                    if (scrollScheduled) return;
                    scrollScheduled = true;
                    requestAnimationFrame(function() {
                        scrollScheduled = false;
                        scrollToBottom(live);
                    });
                }
                function scrollToBottom(live = false) {
                    const el = document.scrollingElement || document.documentElement;
                    const marginPx = 450;
                    let behavior = 'instant';
                    if (live == true) {
                        behavior = 'instant';
                    } else {
                        behavior = 'smooth';
                    }
                    if (isNearBottom(marginPx) || live == false) {
                        el.scrollTo({ top: el.scrollHeight, behavior });
                    }
                    prevScroll = el.scrollHeight;
                }
                function appendToInput(content) {
                    const element = els.appendInput || document.getElementById('_append_input_');
                    if (element) {
                        element.insertAdjacentHTML('beforeend', content);
                        highlightCode(true, element);
                        scheduleScroll();
                    }
                }
                function getStreamContainer() {
                    if (domOutputStream && domOutputStream.isConnected) {
                        return domOutputStream;
                    }
                    let element = els.appendOutput || document.getElementById('_append_output_');
                    if (element) {
                        domOutputStream = element;
                    }
                    return element;
                }
                function appendNode(content) {
                    if (DEBUG_MODE) {
                        log("APPEND NODE: {" + content + "}");
                    }
                    clearStreamBefore();
                    prevScroll = 0;
                    const element = els.nodes || document.getElementById('_nodes_');
                    if (element) {
                        element.classList.remove('empty_list');
                        element.insertAdjacentHTML('beforeend', content);
                        highlightCode(true, element);
                        scheduleScroll();
                    }
                }
                function clean() {
                    if (DEBUG_MODE) {
                        log("-- CLEAN DOM --");
                    }
                    const el = els.nodes || document.getElementById('_nodes_');
                    if (el) {
                        el.replaceChildren();
                    }
                    resetEphemeralDomRefs();
                    els = {};
                    try {
                        if (window.gc) {
                            window.gc();
                        }
                    } catch (e) {
                        // gc not available
                    }
                }
                function appendExtra(id, content) {
                    hideTips();
                    prevScroll = 0;
                    const element = document.getElementById('msg-bot-' + id);
                    if (element) {
                        const extra = element.querySelector('.msg-extra');
                        if (extra) {
                            extra.insertAdjacentHTML('beforeend', content);
                            highlightCode(true, extra);
                            scheduleScroll();
                        }
                    }
                }
                function removeNode(id) {
                    prevScroll = 0;
                    let element = document.getElementById('msg-user-' + id);
                    if (element) {
                        element.remove();
                    }
                    element = document.getElementById('msg-bot-' + id);
                    if (element) {
                        element.remove();
                    }
                    resetEphemeralDomRefs();
                    highlightCode();
                    scheduleScroll();
                }
                function removeNodesFromId(id) {
                    prevScroll = 0;
                    const container = els.nodes || document.getElementById('_nodes_');
                    if (container) {
                        const elements = container.querySelectorAll('.msg-box');
                        let remove = false;
                        elements.forEach(function(element) {
                            if (element.id && element.id.endsWith('-' + id)) {
                                remove = true;
                            }
                            if (remove) {
                                element.remove();
                            }
                        });
                        resetEphemeralDomRefs();
                        highlightCode(true, container);
                        scheduleScroll();
                    }
                }
                function clearStream() {
                    hideTips();
                    if (DEBUG_MODE) {
                        log("STREAM CLEAR");
                    }
                    domLastParagraphBlock = null;
                    domLastCodeBlock = null;
                    domOutputStream = null;
                    const element = getStreamContainer();
                    if (element) {
                        let box = element.querySelector('.msg-box');
                        let msg;
                        if (!box) {
                            box = document.createElement('div');
                            box.classList.add('msg-box');
                            box.classList.add('msg-bot');
                            msg = document.createElement('div');
                            msg.classList.add('msg');
                            box.appendChild(msg);
                            element.appendChild(box);
                        } else {
                            msg = box.querySelector('.msg');
                        }
                        if (msg) {
                            msg.replaceChildren();
                        }
                    }
                }
                function beginStream() {
                    hideTips();
                    if (DEBUG_MODE) {
                        log("STREAM BEGIN");
                    }
                    clearOutput();
                    scheduleScroll();
                }
                function endStream() {
                    if (DEBUG_MODE) {
                        log("STREAM END");
                    }
                    clearOutput();
                }
                function enqueueStream(name_header, content, chunk, replace = false, is_code_block = false) {
                  streamQ.push({name_header, content, chunk, replace, is_code_block});
                  if (!streamRAF) {
                    streamRAF = requestAnimationFrame(drainStream);
                  }
                }                
                function drainStream() {
                  streamRAF = 0;
                  while (streamQ.length) {
                    const {name_header, content, chunk, replace, is_code_block} = streamQ.shift();
                    appendStream(name_header, content, chunk, replace, is_code_block);
                  }
                }
                function appendStream(name_header, content, chunk, replace = false, is_code_block = false) {
                    dropIfDetached(); // clear references to detached elements
                    hideTips();
                    if (DEBUG_MODE) {
                        log("APPEND CHUNK: {" + chunk + "}, CONTENT: {"+content+"}, replace: " + replace + ", is_code_block: " + is_code_block);
                    }
                    const element = getStreamContainer();
                    let msg;
                    if (element) {
                        let box = element.querySelector('.msg-box');
                        if (!box) {
                            box = document.createElement('div');
                            box.classList.add('msg-box');
                            box.classList.add('msg-bot');
                            if (name_header != '') {
                                const name = document.createElement('div');
                                name.classList.add('name-header');
                                name.classList.add('name-bot');
                                name.innerHTML = name_header;
                                box.appendChild(name);
                            }
                            msg = document.createElement('div');
                            msg.classList.add('msg');
                            box.appendChild(msg);
                            element.appendChild(box);
                        } else {
                            msg = box.querySelector('.msg');
                        }
                        if (msg) {
                            if (replace) {
                                msg.replaceChildren();
                                if (content) {
                                  msg.insertAdjacentHTML('afterbegin', content);
                                }
                                let doMath = true;
                                if (is_code_block) {
                                    doMath = false;
                                }
                                highlightCode(doMath, msg);
                                domLastCodeBlock = null;
                                domLastParagraphBlock = null;
                            } else {
                                if (is_code_block) {
                                    let lastCodeBlock;
                                    if (domLastCodeBlock) {
                                        lastCodeBlock = domLastCodeBlock;
                                    } else {
                                        const msgBlocks = msg.querySelectorAll('pre');
                                        if (msgBlocks.length > 0) {
                                            lastCodeBlock = msgBlocks[msgBlocks.length - 1].querySelector('code');
                                        }
                                    }
                                    if (lastCodeBlock) {
                                        lastCodeBlock.insertAdjacentHTML('beforeend', chunk);
                                        domLastCodeBlock = lastCodeBlock;
                                    } else {
                                        msg.insertAdjacentHTML('beforeend', chunk);
                                        domLastCodeBlock = null;
                                    }
                                } else {
                                    domLastCodeBlock = null;
                                    let p = (domLastParagraphBlock && msg.contains(domLastParagraphBlock))
                                        ? domLastParagraphBlock
                                        : (msg.lastElementChild && msg.lastElementChild.tagName === 'P'
                                            ? msg.lastElementChild
                                            : null);
                                    if (p) {
                                        p.insertAdjacentHTML('beforeend', chunk);
                                        domLastParagraphBlock = p;
                                    } else {
                                        msg.insertAdjacentHTML('beforeend', chunk);
                                        const last = msg.lastElementChild;
                                        domLastParagraphBlock = (last && last.tagName === 'P') ? last : null;
                                    }
                                }
                            }
                        }
                    }
                    scheduleScroll(true);
                }
                function nextStream() {
                    hideTips();
                    const element = els.appendOutput || document.getElementById('_append_output_');
                    const elementBefore = els.appendOutputBefore || document.getElementById('_append_output_before_');
                    if (element && elementBefore) {
                        const frag = document.createDocumentFragment();
                        while (element.firstChild) {
                            frag.appendChild(element.firstChild);
                        }
                        elementBefore.appendChild(frag);
                        domLastCodeBlock = null;
                        domLastParagraphBlock = null;
                        scheduleScroll();
                    }
                }
                function clearStreamBefore() {
                    hideTips();
                    const element = els.appendOutputBefore || document.getElementById('_append_output_before_');
                    if (element) {
                        element.replaceChildren();
                    }
                }
                function appendToolOutput(content) {
                    hideToolOutputLoader();
                    enableToolOutput();
                    const elements = document.querySelectorAll('.tool-output');
                    if (elements.length > 0) {
                        const last = elements[elements.length - 1];
                        const contentEl = last.querySelector('.content');
                        if (contentEl) {
                            contentEl.insertAdjacentHTML('beforeend', content);
                        }
                    }
                }
                function updateToolOutput(content) {
                    hideToolOutputLoader();
                    enableToolOutput();
                    const elements = document.querySelectorAll('.tool-output');
                    if (elements.length > 0) {
                        const last = elements[elements.length - 1];
                        const contentEl = last.querySelector('.content');
                        if (contentEl) {
                            contentEl.innerHTML = content;
                        }
                    }
                }
                function clearToolOutput(content) {
                    hideToolOutputLoader();
                    enableToolOutput();
                    const elements = document.querySelectorAll('.tool-output');
                    if (elements.length > 0) {
                        const last = elements[elements.length - 1];
                        const contentEl = last.querySelector('.content');
                        if (contentEl) {
                            contentEl.replaceChildren();
                        }
                    }
                }
                function showToolOutputLoader() {
                    return;
                }
                function hideToolOutputLoader() {
                    const elements = document.querySelectorAll('.msg-bot');
                    if (elements.length > 0) {
                        elements.forEach(function(element) {
                            const contentEl = element.querySelector('.spinner');
                            if (contentEl) {
                                contentEl.style.display = 'none';
                            }
                        });
                    }
                }
                function beginToolOutput() {
                    showToolOutputLoader();
                }
                function endToolOutput() {
                    hideToolOutputLoader();
                }
                function enableToolOutput() {
                    const elements = document.querySelectorAll('.tool-output');
                    if (elements.length > 0) {
                        const last = elements[elements.length - 1];
                        last.style.display = 'block';
                    }
                }
                function disableToolOutput() {
                    const elements = document.querySelectorAll('.tool-output');
                    if (elements.length > 0) {
                        const last = elements[elements.length - 1];
                        last.style.display = 'none';
                    }
                }
                function toggleToolOutput(id) {
                    const element = document.getElementById('msg-bot-' + id);
                    if (element) {
                        const outputEl = element.querySelector('.tool-output');
                        if (outputEl) {
                            const contentEl = outputEl.querySelector('.content');
                            if (contentEl.style.display === 'none') {
                                contentEl.style.display = 'block';
                            } else {
                                contentEl.style.display = 'none';
                            }
                            const toggleEl = outputEl.querySelector('.toggle-cmd-output img');
                            if (toggleEl) {
                                toggleEl.classList.toggle('toggle-expanded');
                            }
                        }
                    }
                }
                function replaceLive(content) {
                    const element = els.appendLive || document.getElementById('_append_live_');
                    if (element) {
                        if (element.classList.contains('hidden')) {
                            element.classList.remove('hidden');
                            element.classList.add('visible');
                        }
                        element.innerHTML = content;
                        highlightCode(true, element);
                        scheduleScroll();
                    }
                }
                function updateFooter(content) {
                    const element = els.footer || document.getElementById('_footer_');
                    if (element) {
                        element.innerHTML = content;
                    }
                }
                function clearNodes() {
                    prevScroll = 0;
                    clearStreamBefore();
                    const element = els.nodes || document.getElementById('_nodes_');
                    if (element) {
                        element.replaceChildren();
                        element.classList.add('empty_list');
                    }
                    resetEphemeralDomRefs();
                }
                function clearInput() {
                    const element = els.appendInput || document.getElementById('_append_input_');
                    if (element) {
                        element.replaceChildren();
                    }
                }
                function clearOutput() {
                    clearStreamBefore();
                    domLastCodeBlock = null;
                    domLastParagraphBlock = null;
                    const element = els.appendOutput || document.getElementById('_append_output_');
                    if (element) {
                        element.replaceChildren();
                    }
                }
                function clearLive() {
                    const element = els.appendLive || document.getElementById('_append_live_');
                    if (element) {
                        if (element.classList.contains('visible')) {
                            element.classList.remove('visible');
                            element.classList.add('hidden');
                            setTimeout(function() {
                                element.replaceChildren();
                                resetEphemeralDomRefs();
                            }, 1000);
                        } else {
                            element.replaceChildren();
                            resetEphemeralDomRefs();
                        }
                    }
                }
                function enableEditIcons() {
                    const container = document.body;
                    if (container) {
                        container.classList.add('display-edit-icons');
                    }
                }
                function disableEditIcons() {
                    const container = document.body;
                    if (container) {
                        container.classList.remove('display-edit-icons');
                    }
                }
                function enableTimestamp() {
                    const container = document.body;
                    if (container) {
                        container.classList.add('display-timestamp');
                    }
                }
                function disableTimestamp() {
                    const container = document.body;
                    if (container) {
                        container.classList.remove('display-timestamp');
                    }
                }
                function enableBlocks() {
                    const container = document.body;
                    if (container) {
                        container.classList.add('display-blocks');
                    }
                }
                function disableBlocks() {
                    const container = document.body;
                    if (container) {
                        container.classList.remove('display-blocks');
                    }
                }
                function updateCSS(styles) {
                    let style = document.getElementById('app-style');
                    if (!style) {
                        style = document.createElement('style');
                        style.id = 'app-style';
                        document.head.appendChild(style);
                    }
                    style.textContent = styles;
                }
                function restoreCollapsedCode(root) {
                    const scope = root || document;
                    const codeWrappers = scope.querySelectorAll('.code-wrapper');
                    codeWrappers.forEach(function(wrapper) {
                        const index = wrapper.getAttribute('data-index');
                        const localeCollapse = wrapper.getAttribute('data-locale-collapse');
                        const localeExpand = wrapper.getAttribute('data-locale-expand');
                        const source = wrapper.querySelector('code');
                        if (source && collapsed_idx.includes(index)) {
                            source.style.display = 'none';
                            const collapseBtn = wrapper.querySelector('.code-header-collapse');
                            if (collapseBtn) {
                                const collapseSpan = collapseBtn.querySelector('span');
                                if (collapseSpan) {
                                    collapseSpan.textContent = localeExpand;
                                }
                                collapseBtn.classList.add('collapsed');
                            }
                        } else if (source) {
                            source.style.display = 'block';
                            const collapseBtn = wrapper.querySelector('.code-header-collapse');
                            if (collapseBtn) {
                                const collapseSpan = collapseBtn.querySelector('span');
                                if (collapseSpan) {
                                    collapseSpan.textContent = localeCollapse;
                                }
                                collapseBtn.classList.remove('collapsed');
                            }
                        }
                    });
                }
                function bridgeCopyCode(text) {
                    if (bridge) {
                        bridge.copy_text(text);
                    }
                }
                function bridgePreviewCode(text) {
                    if (bridge) {
                        bridge.preview_text(text);
                    }
                }
                function bridgeRunCode(text) {
                    if (bridge) {
                        bridge.run_text(text);
                    }
                }
                function bridgeUpdateScrollPosition(pos) {
                    if (bridge) {
                        bridge.update_scroll_position(pos);
                    }
                }
                function getScrollPosition() {
                    bridgeUpdateScrollPosition(window.scrollY);
                }
                function setScrollPosition(pos) {
                    window.scrollTo(0, pos);
                    prevScroll = parseInt(pos);
                }
                function showLoading() {
                    hideTips();
                    const el = els.loader || document.getElementById('_loader_');
                    if (el) {
                        if (el.classList.contains('hidden')) {
                            el.classList.remove('hidden');
                        }
                        el.classList.add('visible');
                    }
                }
                function hideLoading() {
                    const el = els.loader || document.getElementById('_loader_');
                    if (el) {
                        if (el.classList.contains('visible')) {
                            el.classList.remove('visible');
                        }
                        el.classList.add('hidden');
                    }
                }
                document.addEventListener('DOMContentLoaded', function() {
                    new QWebChannel(qt.webChannelTransport, function (channel) {
                        bridge = channel.objects.bridge;
                        bridge.chunk.connect((name, html, chunk, replace, isCode) => {
                            appendStream(name, html, chunk, replace, isCode);
                        });
                        if (bridge.js_ready) bridge.js_ready();
                    });
                    initDomRefs();
                    const container = els.container;
                    function addClassToMsg(id, className) {
                        const msgElement = document.getElementById('msg-bot-' + id);
                        if (msgElement) {
                            msgElement.classList.add(className);
                        }
                    }
                    function removeClassFromMsg(id, className) {
                        const msgElement = document.getElementById('msg-bot-' + id);
                        if (msgElement) {
                            msgElement.classList.remove(className);
                        }
                    }
                    container.addEventListener('mouseover', function(event) {
                        if (event.target.classList.contains('action-img')) {
                            const id = event.target.getAttribute('data-id');
                            addClassToMsg(id, 'msg-highlight');
                        }
                    });
                    container.addEventListener('mouseout', function(event) {
                        if (event.target.classList.contains('action-img')) {
                            const id = event.target.getAttribute('data-id');
                            removeClassFromMsg(id, 'msg-highlight');
                        }
                    });
                    container.addEventListener('click', function(event) {
                        const copyButton = event.target.closest('.code-header-copy');
                        if (copyButton) {
                            event.preventDefault();
                            const parent = event.target.closest('.code-wrapper');
                            const source = parent.querySelector('code');
                            const localeCopy = parent.getAttribute('data-locale-copy');
                            const localeCopied = parent.getAttribute('data-locale-copied');
                            if (source) {
                                const text = source.textContent || source.innerText;
                                bridgeCopyCode(text);
                                const copySpan = copyButton.querySelector('span');
                                if (copySpan) {
                                    copySpan.textContent = localeCopied;
                                    setTimeout(function() {
                                        copySpan.textContent = localeCopy;
                                    }, 1000);
                                }
                            }
                        }
                        const runButton = event.target.closest('.code-header-run');
                        if (runButton) {
                            event.preventDefault();
                            const parent = event.target.closest('.code-wrapper');
                            const source = parent.querySelector('code');
                            if (source) {
                                const text = source.textContent || source.innerText;
                                bridgeRunCode(text);
                            }
                        }
                        const previewButton = event.target.closest('.code-header-preview');
                        if (previewButton) {
                            event.preventDefault();
                            const parent = event.target.closest('.code-wrapper');
                            const source = parent.querySelector('code');
                            if (source) {
                                const text = source.textContent || source.innerText;
                                bridgePreviewCode(text);
                            }
                        }
                        const collapseButton = event.target.closest('.code-header-collapse');
                        if (collapseButton) {
                            event.preventDefault();
                            const parent = collapseButton.closest('.code-wrapper');
                            const index = parent.getAttribute('data-index');
                            const source = parent.querySelector('code');
                            const localeCollapse = parent.getAttribute('data-locale-collapse');
                            const localeExpand = parent.getAttribute('data-locale-expand');
                            if (source) {
                                if (source.style.display === 'none') {
                                    source.style.display = 'block';
                                    collapseButton.classList.remove('collapsed');
                                    const idx = collapsed_idx.indexOf(index);
                                    if (idx !== -1) {
                                        collapsed_idx.splice(idx, 1);
                                        const collapseSpan = collapseButton.querySelector('span');
                                        if (collapseSpan) {
                                            collapseSpan.textContent = localeCollapse;
                                        }
                                    }
                                } else {
                                    source.style.display = 'none';
                                    collapseButton.classList.add('collapsed');
                                    if (!collapsed_idx.includes(index)) {
                                        collapsed_idx.push(index);
                                        const collapseSpan = collapseButton.querySelector('span');
                                        if (collapseSpan) {
                                            collapseSpan.textContent = localeExpand;
                                        }
                                    }
                                }
                            }
                        }
                    });
                });
                setTimeout(cycleTips, 60000);  // after 60 seconds
                </script>
            </head>
            <body """
    _HTML_P4 = """>
            <div id="container">
                <div id="_nodes_" class="nodes empty_list"></div>
                <div id="_append_input_" class="append_input"></div>
                <div id="_append_output_before_" class="append_output"></div>
                <div id="_append_output_" class="append_output"></div>
                <div id="_append_live_" class="append_live hidden"></div>
                <div id="_footer_" class="footer"></div>
                <div id="_loader_" class="loader-global hidden">
                    <div class="lds-ring"><div></div><div></div><div></div><div></div></div>
                </div>
                <div id="tips" class="tips"></div>
            </div>
            </body>
            </html>
            """

    _SPINNER = """
        .lds-ring {
          /* change color here */
          color: #1c4c5b
        }
        .lds-ring,
        .lds-ring div {
          box-sizing: border-box;
        }
        .lds-ring {
          display: inline-block;
          position: relative;
          width: 80px;
          height: 80px;
        }
        .lds-ring div {
          box-sizing: border-box;
          display: block;
          position: absolute;
          width: 64px;
          height: 64px;
          margin: 8px;
          border: 8px solid currentColor;
          border-radius: 50%;
          animation: lds-ring 1.2s cubic-bezier(0.5, 0, 0.5, 1) infinite;
          border-color: currentColor transparent transparent transparent;
        }
        .lds-ring div:nth-child(1) {
          animation-delay: -0.45s;
        }
        .lds-ring div:nth-child(2) {
          animation-delay: -0.3s;
        }
        .lds-ring div:nth-child(3) {
          animation-delay: -0.15s;
        }
        @keyframes lds-ring {
          0% {
            transform: rotate(0deg);
          }
          100% {
            transform: rotate(360deg);
          }
        }
        """

    def __init__(self, window=None):
        """
        HTML Body

        :param window: Window instance
        """
        self.window = window
        self.highlight = SyntaxHighlight(window)
        self._tip_keys = tuple(f"output.tips.{i}" for i in range(1, self.NUM_TIPS + 1))
        self._syntax_dark = (
            "dracula",
            "fruity",
            "github-dark",
            "gruvbox-dark",
            "inkpot",
            "material",
            "monokai",
            "native",
            "nord",
            "nord-darker",
            "one-dark",
            "paraiso-dark",
            "rrt",
            "solarized-dark",
            "stata-dark",
            "vim",
            "zenburn",
        )


    def is_timestamp_enabled(self) -> bool:
        """
        Check if timestamp is enabled

        :return: True if timestamp is enabled
        """
        return self.window.core.config.get('output_timestamp')

    def prepare_styles(self) -> str:
        """
        Prepare CSS styles

        :return: CSS styles
        """
        cfg = self.window.core.config
        fonts_path = os.path.join(cfg.get_app_path(), "data", "fonts").replace("\\", "/")
        syntax_style = self.window.core.config.get("render.code_syntax") or "default"

        theme_css = self.window.controller.theme.markdown.get_web_css().replace('%fonts%', fonts_path)
        parts = [self._SPINNER, theme_css,
                 "pre { color: #fff; }" if syntax_style in self._syntax_dark else "pre { color: #000; }",
                 self.highlight.get_style_defs()]
        return "\n".join(parts)

    def prepare_action_icons(self, ctx: CtxItem) -> str:
        """
        Append action icons

        :param ctx: context item
        :return: HTML code
        """
        icons_html = "".join(self.get_action_icons(ctx, all=True))
        if icons_html:
            return f'<div class="action-icons" data-id="{ctx.id}">{icons_html}</div>'
        return ""

    def get_action_icons(self, ctx: CtxItem, all: bool = False) -> List[str]:
        """
        Get action icons for context item

        :param ctx: context item
        :param all: True to show all icons
        :return: list of icons
        """
        icons: List[str] = []
        if ctx.output:
            cid = ctx.id
            t = trans
            icons.append(
                f'<a href="extra-audio-read:{cid}" class="action-icon" data-id="{cid}" role="button"><span class="cmd">{self.get_icon("volume", t("ctx.extra.audio"), ctx)}</span></a>')
            icons.append(
                f'<a href="extra-copy:{cid}" class="action-icon" data-id="{cid}" role="button"><span class="cmd">{self.get_icon("copy", t("ctx.extra.copy"), ctx)}</span></a>')
            icons.append(
                f'<a href="extra-replay:{cid}" class="action-icon" data-id="{cid}" role="button"><span class="cmd">{self.get_icon("reload", t("ctx.extra.reply"), ctx)}</span></a>')
            icons.append(
                f'<a href="extra-edit:{cid}" class="action-icon edit-icon" data-id="{cid}" role="button"><span class="cmd">{self.get_icon("edit", t("ctx.extra.edit"), ctx)}</span></a>')
            icons.append(
                f'<a href="extra-delete:{cid}" class="action-icon edit-icon" data-id="{cid}" role="button"><span class="cmd">{self.get_icon("delete", t("ctx.extra.delete"), ctx)}</span></a>')
            if not self.window.core.ctx.is_first_item(cid):
                icons.append(
                    f'<a href="extra-join:{cid}" class="action-icon edit-icon" data-id="{cid}" role="button"><span class="cmd">{self.get_icon("playlist_add", t("ctx.extra.join"), ctx)}</span></a>')
        return icons

    def get_icon(
            self,
            icon: str,
            title: Optional[str] = None,
            item: Optional[CtxItem] = None
    ) -> str:
        """
        Get icon

        :param icon: icon name
        :param title: icon title
        :param item: context item
        :return: icon HTML
        """
        app_path = self.window.core.config.get_app_path()
        icon_path = os.path.join(app_path, "data", "icons", f"{icon}.svg")
        return f'<img src="file://{icon_path}" class="action-img" title="{title}" alt="{title}" data-id="{item.id}">'

    def get_image_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get image HTML

        :param url: URL to image
        :param num: number of image
        :param num_all: number of all images
        :return: HTML code
        """
        url, path = self.window.core.filesystem.extract_local_url(url)
        basename = os.path.basename(path)
        return f'<div class="extra-src-img-box" title="{url}"><div class="img-outer"><div class="img-wrapper"><a href="{url}"><img src="{path}" class="image"></a></div><a href="{url}" class="title">{basename}</a></div></div><br/>'

    def get_url_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get URL HTML

        :param url: external URL
        :param num: number of URL
        :param num_all: number of all URLs
        :return: HTML code
        """
        app_path = self.window.core.config.get_app_path()
        icon_path = os.path.join(app_path, "data", "icons", "language.svg").replace("\\", "/")
        icon = f'<img src="file://{icon_path}" class="extra-src-icon">'
        num_str = f" [{num}]" if (num is not None and num_all is not None and num_all > 1) else ""
        return f'{icon}<a href="{url}" title="{url}">{url}</a> <small>{num_str}</small>'

    def get_docs_html(self, docs: List[Dict]) -> str:
        """
        Get Llama-index doc metadata HTML

        :param docs: list of document metadata
        :return: HTML code
        """
        html_parts: List[str] = []
        src_parts: List[str] = []
        num = 1
        limit = 3

        try:
            for item in docs:
                for uuid, doc_json in item.items():
                    entries = ", ".join(f"<b>{k}:</b> {doc_json[k]}" for k in doc_json)
                    src_parts.append(f"<p><small>[{num}] {uuid}: {entries}</small></p>")
                    num += 1
                    if num >= limit:
                        break
                if num >= limit:
                    break
        except Exception:
            pass

        if src_parts:
            app_path = self.window.core.config.get_app_path()
            icon_path = os.path.join(app_path, "data", "icons", "db.svg").replace("\\", "/")
            icon = f'<img src="file://{icon_path}" class="extra-src-icon">'
            html_parts.append(f'<p>{icon}<small><b>{trans("chat.prefix.doc")}:</b></small></p>')
            html_parts.append('<div class="cmd">')
            html_parts.append(f"<p>{''.join(src_parts)}</p>")
            html_parts.append("</div> ")

        return "".join(html_parts)

    def get_file_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get file HTML

        :param url: URL to file
        :param num: number of file
        :param num_all: number of all files
        :return: HTML code
        """
        app_path = self.window.core.config.get_app_path()
        icon_path = os.path.join(app_path, "data", "icons", "attachments.svg").replace("\\", "/")
        icon = f'<img src="file://{icon_path}" class="extra-src-icon">'
        num_str = f" [{num}]" if (num is not None and num_all is not None and num_all > 1) else ""
        url, path = self.window.core.filesystem.extract_local_url(url)
        return f'{icon} <b>{num_str}</b> <a href="{url}">{path}</a>'

    def prepare_tool_extra(self, ctx: CtxItem) -> str:
        """
        Prepare footer extra

        :param ctx: context item
        :return: HTML code
        """
        extra = ctx.extra
        if not extra:
            return ""

        parts: List[str] = ['<div class="msg-extra">']

        if "plugin" in extra:
            event = Event(Event.TOOL_OUTPUT_RENDER, {
                'tool': extra["plugin"],
                'html': '',
                'multiple': False,
                'content': extra,
            })
            event.ctx = ctx
            self.window.dispatch(event, all=True)
            if event.data['html']:
                parts.append(f'<div class="tool-output-block">{event.data["html"]}</div>')

        elif "tool_output" in extra and isinstance(extra["tool_output"], list):
            for tool in extra["tool_output"]:
                if "plugin" not in tool:
                    continue
                event = Event(Event.TOOL_OUTPUT_RENDER, {
                    'tool': tool["plugin"],
                    'html': '',
                    'multiple': True,
                    'content': tool,
                })
                event.ctx = ctx
                self.window.dispatch(event, all=True)
                if event.data['html']:
                    parts.append(f'<div class="tool-output-block">{event.data["html"]}</div>')

        parts.append("</div>")
        return "".join(parts)

    def get_all_tips(self) -> str:
        """
        Get all tips for the output view

        :return: JSON string of tips
        """
        if not self.window.core.config.get("layout.tooltips", False):
            return "[]"

        _trans = trans
        prefix = _trans("output.tips.prefix")
        keys = self._tip_keys if hasattr(self, "_tip_keys") and len(self._tip_keys) == self.NUM_TIPS \
            else tuple(f"output.tips.{i}" for i in range(1, self.NUM_TIPS + 1))

        tips = [f"<p><b>{prefix}</b>: {_trans(k)}</p>" for k in keys]
        _shuffle(tips)
        return _json_dumps(tips)

    def get_html(self, pid: int) -> str:
        """
        Build webview HTML code (fast path, minimal allocations)
        """
        cfg_get = self.window.core.config.get
        style = cfg_get("theme.style", "blocks")
        classes = ["theme-" + style]
        if cfg_get('render.blocks'):
            classes.append("display-blocks")
        if cfg_get('ctx.edit_icons'):
            classes.append("display-edit-icons")
        if self.is_timestamp_enabled():
            classes.append("display-timestamp")
        classes_str = f' class="{" ".join(classes)}"' if classes else ""
        styles_css = self.prepare_styles()
        tips_json = self.get_all_tips()

        return ''.join((
            self._HTML_P0,
            styles_css,
            self._HTML_P1,
            str(pid),
            self._HTML_P2,
            tips_json,
            self._HTML_P3,
            classes_str,
            self._HTML_P4,
        ))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.14 13:00:00                  #
# ================================================== #

import datetime
import os
import shutil
from typing import Any, Optional, Dict

from PySide6.QtWidgets import QVBoxLayout, QWidget, QHBoxLayout

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AGENT_OPENAI,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_IMAGE,
    MODE_RESEARCH,
    MODE_COMPUTER,
)
from pygpt_net.item.preset import PresetItem
from pygpt_net.utils import trans

from .experts import Experts

class Editor:

    TAB_IDX = {
        "general": 0,
        "personalize": 1,
        "experts": 2,
        "remote_tools": 3,
    }

    def __init__(self, window=None):
        """
        Presets editor controller

        :param window: Window instance
        """
        self.window = window
        self.experts = Experts(window)
        self.built = False
        self.tab_options_idx = {}
        self.opened = False
        self.tmp_avatar = None
        self.options = {
            "filename": {
                "type": "text",
                "label": "preset.filename",
            },
            "name": {
                "type": "text",
                "label": "preset.name",
            },
            "ai_name": {
                "type": "text",
                "label": "preset.ai_name",
            },
            "ai_avatar": {
                "type": "text",
                "label": "preset.ai_avatar",
            },
            "ai_personalize": {
                "type": "bool",
                "label": "preset.ai_personalize",
                "description": "preset.ai_personalize.desc",
            },
            "user_name": {
                "type": "text",
                "label": "preset.user_name",
            },
            "description": {
                "type": "text",
                "label": "preset.description",
                "placeholder": "preset.description.desc",
            },
            MODE_IMAGE: {
                "type": "bool",
                "label": "preset.img",
            },
            MODE_CHAT: {
                "type": "bool",
                "label": "preset.chat",
            },
            MODE_COMPLETION: {
                "type": "bool",
                "label": "preset.completion",
            },
            MODE_VISION: {
                "type": "bool",
                "label": "preset.vision",
            },
            #MODE_LANGCHAIN: {
            #    "type": "bool",
            #    "label": "preset.langchain",
            #},
            MODE_EXPERT: {
                "type": "bool",
                "label": "preset.expert",
            },
            MODE_AGENT_LLAMA: {
                "type": "bool",
                "label": "preset.agent_llama",
            },
            MODE_AGENT: {
                "type": "bool",
                "label": "preset.agent",
            },
            MODE_AGENT_OPENAI: {
                "type": "bool",
                "label": "preset.agent_openai",
            },
            MODE_AUDIO: {
                "type": "bool",
                "label": "preset.audio",
            },
            MODE_RESEARCH: {
                "type": "bool",
                "label": "preset.research",
            },
            # "assistant": {
            # "type": "bool",
            # "label": "preset.assistant",
            # },
            MODE_LLAMA_INDEX: {
                "type": "bool",
                "label": "preset.llama_index",
            },
            MODE_COMPUTER: {
                "type": "bool",
                "label": "preset.computer",
            },
            "model": {
                "label": "toolbox.model.label",
                "type": "combo",
                "use": "models",
            },
            "remote_tools": {
                "label": "toolbox.remote_tools.label",
                "type": "bool_list",
                "use": "remote_tools_openai",
            },
            "temperature": {
                "type": "float",
                "slider": True,
                "label": "preset.temperature",
                "min": 0,
                "max": 2,
                "step": 1,
                "multiplier": 100,
            },
            "prompt": {
                "type": "textarea",
                "label": "",
                "context_options": ["prompt.template.paste"],
            },
            "idx": {
                "type": "combo",
                "label": "preset.idx",
                "description": "preset.idx.desc",
                "use": "idx",
            },
            "agent_provider": {
                "type": "combo",
                "label": "preset.agent_provider",
                "description": "preset.agent_provider.desc",
                "use": "agent_provider_llama",
                "extra": {
                    "urls": {
                        "Help": "https://pygpt.readthedocs.io/en/latest/modes.html#agent-llamaindex"
                    },
                }
            },
            "agent_provider_openai": {
                "type": "combo",
                "label": "preset.agent_provider",
                "description": "preset.agent_provider.desc",
                "use": "agent_provider_openai",
                "extra": {
                    "urls": {
                        "Help": "https://pygpt.readthedocs.io/en/latest/modes.html#agent-openai"
                    },
                }
            },
        }
        self.hidden_by_mode = {  # hidden fields by mode
            MODE_CHAT: ["idx"],
            MODE_AGENT_LLAMA: ["temperature"],
            MODE_AGENT_OPENAI: ["temperature"],
        }
        self.id = "preset"
        self.current = None

    def get_options(self) -> Dict[str, Dict[str, Any]]:
        """
        Get preset options

        :return: preset options
        """
        return self.options

    def get_option(self, id: str) -> Dict[str, Any]:
        """
        Get preset option

        :param id: option id
        :return: preset option
        """
        return self.options[id]

    def show_hide_by_mode(self):
        """Show or hide fields by mode"""
        # TODO: implement this!
        return
        mode = self.window.core.config.get('mode')
        for key in self.options:
            if mode in self.hidden_by_mode:
                if key in self.hidden_by_mode[mode]:
                    self.window.ui.config[self.id][key].setVisible(False)
                    label = 'preset.' + key + '.label'
                    if label in self.window.ui.nodes:
                        self.window.ui.nodes[label].setVisible(False)
                else:
                    self.window.ui.config[self.id][key].setVisible(True)
                    label = 'preset.' + key + '.label'
                    if label in self.window.ui.nodes:
                        self.window.ui.nodes[label].setVisible(True)

    def setup(self):
        """Setup preset editor"""
        # update after agents register
        self.append_extra_config()
        self.window.ui.config[self.id]['agent_provider'].set_keys(
            self.window.controller.config.placeholder.apply_by_id('agent_provider_llama')
        )
        self.window.ui.config[self.id]['agent_provider_openai'].set_keys(
            self.window.controller.config.placeholder.apply_by_id('agent_provider_openai')
        )

        # add hooks for config update in real-time
        self.window.ui.add_hook("update.preset.prompt", self.hook_update)
        self.window.ui.add_hook("update.preset.agent_provider", self.hook_update)
        self.window.ui.add_hook("update.preset.agent_provider_openai", self.hook_update)

    def toggle_extra_options(self):
        """Toggle extra options in preset editor"""
        if not self.tab_options_idx:
            return
        mode = self.window.core.config.get('mode')
        if mode not in [MODE_AGENT_OPENAI, MODE_AGENT_LLAMA]:
            # show base prompt
            self.window.ui.tabs['preset.editor.extra'].setTabVisible(0, True)
            # hide all tabs
            for id in self.tab_options_idx:
                for tab_idx in self.tab_options_idx[id]:
                    if self.window.ui.tabs['preset.editor.extra'].count() >= tab_idx:
                        self.window.ui.tabs['preset.editor.extra'].setTabVisible(tab_idx, False)
            return
        else:
            # hide all tabs
            for id in self.tab_options_idx:
                for tab_idx in self.tab_options_idx[id]:
                    if self.window.ui.tabs['preset.editor.extra'].count() >= tab_idx:
                        self.window.ui.tabs['preset.editor.extra'].setTabVisible(tab_idx, False)

            self.toggle_extra_options_by_provider()

    def toggle_extra_options_by_provider(self):
        """Toggle extra options in preset editor by provider"""
        if not self.tab_options_idx:
            # show base prompt
            self.window.ui.tabs['preset.editor.extra'].setTabVisible(0, True)
            return

        mode = self.window.core.config.get('mode')
        key_agent = ""

        if mode in [MODE_AGENT_OPENAI, MODE_AGENT_LLAMA]:
            # get current provider
            if mode == MODE_AGENT_LLAMA:
                key_agent = "agent_provider"
            elif mode == MODE_AGENT_OPENAI:
                key_agent = "agent_provider_openai"

            current_provider = self.window.controller.config.get_value(
                parent_id=self.id,
                key=key_agent,
                option=self.options[key_agent],
            )
            if current_provider is None or current_provider == "":
                # show base prompt
                self.window.ui.tabs['preset.editor.extra'].setTabVisible(0, True)
                return

            # show all tabs for current provider
            for id in self.tab_options_idx:

                self.window.ui.tabs['preset.editor.extra'].setTabVisible(0, False)

                if id != current_provider:
                    for tab_idx in self.tab_options_idx[id]:
                        if self.window.ui.tabs['preset.editor.extra'].count() >= tab_idx:
                            self.window.ui.tabs['preset.editor.extra'].setTabVisible(tab_idx, False)
                else:
                    for tab_idx in self.tab_options_idx[id]:
                        if self.window.ui.tabs['preset.editor.extra'].count() >= tab_idx:
                            self.window.ui.tabs['preset.editor.extra'].setTabVisible(tab_idx, True)

            # show base prompt if no custom options in current agent
            agent = self.window.core.agents.provider.get(current_provider)
            if not agent:
                self.window.ui.tabs['preset.editor.extra'].setTabVisible(0, True)
                return
            option_tabs = agent.get_options()
            if not option_tabs or len(option_tabs) == 0:
                self.window.ui.tabs['preset.editor.extra'].setTabVisible(0, True)

    def load_extra_options(self, preset: PresetItem):
        """
        Load extra options for preset editor

        :param preset: preset item
        """
        mode = self.window.core.config.get('mode')
        id = None
        if mode == MODE_AGENT_OPENAI:
            if preset.agent_provider_openai is None or preset.agent_provider_openai == "":
                return
            id = preset.agent_provider_openai
        elif mode == MODE_AGENT_LLAMA:
            if preset.agent_provider is None or preset.agent_provider == "":
                return
            id = preset.agent_provider
        else:
            return

        # update options in UI
        agent = self.window.core.agents.provider.get(id)
        if not agent:
            return
        if not preset.extra or id not in preset.extra:
            return

        data_dict = preset.extra[id]
        option_tabs = agent.get_options()
        for option_tab_id in data_dict:
            parent_key = ""
            if mode == MODE_AGENT_OPENAI:
                parent_key = preset.agent_provider_openai
            elif mode == MODE_AGENT_LLAMA:
                parent_key = preset.agent_provider
            option_key = "agent." + parent_key + "." + option_tab_id
            if option_key not in self.window.ui.config:
                continue

            extra_options = option_tabs.get(option_tab_id, {}).get('options', {})
            for key in extra_options:
                value = data_dict[option_tab_id].get(key, None)
                if value is not None:
                    self.window.controller.config.apply_value(
                        parent_id=option_key,
                        key=key,
                        option=extra_options[key],
                        value=value,
                    )
                else:
                    # from defaults
                    if "default" not in extra_options[key]:
                        continue
                    self.window.controller.config.apply_value(
                        parent_id=option_key,
                        key=key,
                        option=extra_options[key],
                        value=extra_options[key].get('default', None),
                    )

    def load_extra_defaults(self):
        """Load extra options defaults for preset editor"""
        if not self.tab_options_idx:
            return
        mode = self.window.core.config.get('mode')
        if mode not in [MODE_AGENT_OPENAI, MODE_AGENT_LLAMA]:
            return

        # load defaults for all tabs
        for id in self.tab_options_idx:
            agent = self.window.core.agents.provider.get(id)
            if not agent:
                continue
            option_tabs = agent.get_options()
            for option_tab_id in option_tabs:
                option_key = "agent." + id + "." + option_tab_id
                if option_key not in self.window.ui.config:
                    continue
                extra_options = option_tabs[option_tab_id]['options']
                for key in extra_options:
                    value = extra_options[key].get('default', None)
                    if value is not None:
                        self.window.controller.config.apply_value(
                            parent_id=option_key,
                            key=key,
                            option=extra_options[key],
                            value=value,
                        )

    def load_extra_defaults_current(self):
        """Load extra options defaults on mode change"""
        if not self.tab_options_idx:
            return

        if not self.current:
            return

        mode = self.window.core.config.get('mode')
        if mode not in [MODE_AGENT_OPENAI, MODE_AGENT_LLAMA]:
            return

        preset = self.window.core.presets.get_by_uuid(self.current)
        if not preset:
            return

        current_provider_id = None
        if mode == MODE_AGENT_OPENAI:
            current_provider_id = preset.agent_provider_openai if preset else None
        elif mode == MODE_AGENT_LLAMA:
            current_provider_id = preset.agent_provider if preset else None

        # load defaults for all tabs
        for id in self.tab_options_idx:
            # skip current provider
            if current_provider_id and id == current_provider_id:
                continue
            agent = self.window.core.agents.provider.get(id)
            if not agent:
                continue
            option_tabs = agent.get_options()
            for option_tab_id in option_tabs:
                option_key = "agent." + id + "." + option_tab_id
                if option_key not in self.window.ui.config:
                    continue
                extra_options = option_tabs[option_tab_id]['options']
                for key in extra_options:
                    value = extra_options[key].get('default', None)
                    if value is not None:
                        # check current, apply only if current is empty
                        current_value = self.window.controller.config.get_value(
                            parent_id=option_key,
                            key=key,
                            option=extra_options[key],
                        )
                        if current_value is not None and current_value != "":
                            continue

                        self.window.controller.config.apply_value(
                            parent_id=option_key,
                            key=key,
                            option=extra_options[key],
                            value=value,
                        )

    def append_extra_options(self, preset: PresetItem):
        """
        Append extra options for preset editor

        :param preset: preset item
        """
        mode = self.window.core.config.get('mode')
        exclude_ids = [
            "__prompt__",
        ]
        id = None
        if mode == MODE_AGENT_OPENAI:
            id = preset.agent_provider_openai
        elif mode == MODE_AGENT_LLAMA:
            id = preset.agent_provider
        else:
            return

        options = {}
        agent = self.window.core.agents.provider.get(id)
        if not agent:
            return options
        option_tabs = agent.get_options()
        if not option_tabs:
            return options
        data_dict = {}
        for option_tab_id in option_tabs:
            if option_tab_id in exclude_ids:
                continue
            option_key = "agent." + id + "." + option_tab_id
            if option_key not in self.window.ui.config:
                continue
            if option_tab_id not in data_dict:
                data_dict[option_tab_id] = {}
            extra_options = option_tabs[option_tab_id]['options']
            for key in extra_options:
                data_dict[option_tab_id][key] = self.window.controller.config.get_value(
                    parent_id=option_key,
                    key=key,
                    option=extra_options[key],
                )
        if preset.extra is None:
            preset.extra = {}
        preset.extra[id] = data_dict

    def append_extra_config(self):
        """Build extra configuration for the preset editor dialog"""
        if self.built:
            return

        exclude_ids = [
            "__prompt__",
        ]
        agents = self.window.core.agents.provider.all()
        tabs = self.window.ui.tabs['preset.editor.extra']
        tab_idx = 1
        for id in agents:
            agent = agents[id]
            option_tabs = agent.get_options()
            if not option_tabs:
                continue

            # append tabs
            for option_tab_id in option_tabs:
                if option_tab_id in exclude_ids:
                    continue

                option = option_tabs[option_tab_id]
                title = option.get('label', '')
                config_id = "agent." + id + "." + option_tab_id
                widgets, options = self.window.ui.dialogs.preset.build_option_widgets(config_id, option['options'])
                layout = QVBoxLayout()
                layout.setContentsMargins(0, 10, 0, 10)

                checkboxLayout = QHBoxLayout()
                for key in options:
                    opt_layout = options[key]
                    if option['options'][key]['type'] == 'bool':
                        # checkbox
                        checkboxLayout.addLayout(opt_layout)
                    else:
                        layout.addLayout(opt_layout)
                layout.addStretch(1)
                layout.addLayout(checkboxLayout)

                # as tab
                tab_widget = QWidget()
                tab_widget.setLayout(layout)
                tabs.addTab(tab_widget, title)

                # store mapping: agent id -> [tab index]
                if id not in self.tab_options_idx:
                    self.tab_options_idx[id] = []
                if tab_idx not in self.tab_options_idx[id]:
                    self.tab_options_idx[id].append(tab_idx)
                tab_idx += 1

        self.built = True

    def append_default_prompt(self):
        """Append default prompt to the preset editor"""
        mode = self.window.core.config.get('mode')
        if mode not in [MODE_AGENT_OPENAI, MODE_AGENT_LLAMA]:
            return

        parent_key = ""
        if mode == MODE_AGENT_OPENAI:
            parent_key = "agent_provider_openai"
        elif mode == MODE_AGENT_LLAMA:
            parent_key = "agent_provider"

        # get current provider
        current_provider = self.window.controller.config.get_value(
            parent_id=self.id,
            key=parent_key,
            option=self.options[parent_key],
        )
        if current_provider is None or current_provider == "":
            return

        # get agent by provider
        agent = self.window.core.agents.provider.get(current_provider)
        if not agent:
            return

        # get default prompt
        default_prompt = agent.get_default_prompt()
        if default_prompt is None or default_prompt == "":
            return

        # set default prompt to the prompt field
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="prompt",
            option=self.options["prompt"],
            value=default_prompt,
        )

    def hook_update(
            self,
            key: str,
            value: Any,
            caller,
            *args,
            **kwargs
    ):
        """
        Hook: on settings update in real-time (prompt)

        :param key: setting key
        :param value: setting value
        :param caller: caller id
        :param args: additional arguments
        :param kwargs: additional keyword arguments
        """
        mode = self.window.core.config.get('mode')
        if key == "prompt":
            self.window.core.config.set('prompt', value)
            if mode == MODE_ASSISTANT:
                self.window.controller.assistant.from_global()  # update current assistant, never called!!!!!
            else:
                self.window.controller.presets.from_global()  # update current preset

        # show/hide extra options
        elif key in ["agent_provider_openai", "agent_provider"]:
            self.toggle_extra_options_by_provider()
            self.append_default_prompt()
            self.load_extra_defaults_current()

    def edit(self, idx: Optional[int] = None):
        """
        Open preset editor

        :param idx: preset index (row index)
        """
        preset = None
        if idx is not None:
            mode = self.window.core.config.get('mode')
            preset = self.window.core.presets.get_by_idx(idx, mode)
        self.init(preset)
        self.window.ui.dialogs.open_editor('editor.preset.presets', idx, width=800)

    def init(self, id: Optional[str] = None):
        """
        Initialize preset editor

        :param id: preset id (filename)
        """
        self.opened = True
        data = PresetItem()
        data.name = ""
        data.filename = ""

        if id is None:
            self.experts.update_list()
            self.window.ui.config[self.id]['idx'].set_value("_")  # reset idx combo if new preset

        if id is not None and id != "":
            if id in self.window.core.presets.items:
                data = self.window.core.presets.items[id]
                data.filename = id
                self.current = data.uuid
        else:
            self.load_extra_defaults()

        if data.name is None:
            data.name = ""
        if data.ai_name is None:
            data.ai_name = ""
        if data.user_name is None:
            data.user_name = ""
        if data.prompt is None:
            data.prompt = ""
        if data.filename is None:
            data.filename = ""

        # set current mode at start
        if id is None:
            mode = self.window.core.config.get("mode")
            if mode == MODE_CHAT:
                data.chat = True
            elif mode == MODE_COMPLETION:
                data.completion = True
            elif mode == MODE_IMAGE:
                data.img = True
            elif mode == MODE_VISION:
                data.vision = True
            # elif mode == MODE_LANGCHAIN:
                # data.langchain = True
            # elif mode == MODE_ASSISTANT:
                # data.assistant = True
            elif mode == MODE_LLAMA_INDEX:
                data.llama_index = True
            elif mode == MODE_EXPERT:
                data.expert = True
            elif mode == MODE_AGENT:
                data.agent = True
            elif mode == MODE_AGENT_LLAMA:
                data.agent_llama = True
            elif mode == MODE_AGENT_OPENAI:
                data.agent_openai = True
            elif mode == MODE_AUDIO:
                data.audio = True
            elif mode == MODE_RESEARCH:
                data.research = True
            elif mode == MODE_COMPUTER:
                data.computer = True

        options = {}
        data_dict = data.to_dict()
        for key in self.options:
            options[key] = self.options[key]
            options[key]['value'] = data_dict[key]

        # load options
        self.window.controller.config.load_options(
            self.id,
            options,
        )

        # load extra options
        self.load_extra_options(data)

        # toggle extra options
        self.toggle_extra_options()

        # update experts list, after ID loaded
        self.experts.update_list()

        # setup avatar config
        self.update_avatar_config(data)

        # set focus to name field
        current_model = self.window.core.config.get('model')

        # set current model in combo box as selected
        if id is None:
            self.window.ui.config[self.id]['model'].set_value(current_model)
        self.window.ui.config[self.id]['name'].setFocus()
        self.show_hide_by_mode()
        self.toggle_extra_options_by_provider()
        if id is None:
            self.append_default_prompt()

    def save(
            self,
            force: bool = False,
            close: bool = True
    ):
        """
        Save preset

        :param force: force overwrite file
        :param close: close dialog
        """
        id = self.window.controller.config.get_value(
            parent_id=self.id,
            key="filename",
            option=self.options["filename"],
        )
        mode = self.window.core.config.get("mode")
        modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_IMAGE,
            MODE_VISION,
            # MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_EXPERT,
            MODE_AGENT_LLAMA,
            MODE_AGENT,
            MODE_AGENT_OPENAI,
            MODE_AUDIO,
            MODE_COMPUTER,
        ]

        # disallow editing default preset
        if id == "current." + mode:
            self.window.ui.dialogs.alert("Reserved ID. Please use another ID.")
            return

        if id is None or id == "":
            name = self.window.controller.config.get_value(
                parent_id=self.id,
                key="name",
                option=self.options["name"],
            )
            if name is None or name == "":
                self.window.ui.dialogs.alert(trans('alert.preset.empty_id'))
                self.window.update_status(trans('status.preset.empty_id'))
                return

            # generate new filename
            id = self.window.controller.presets.make_filename(name)
            path = os.path.join(
                self.window.core.config.path,
                "presets",
                id + ".json",
            )
            if os.path.exists(path) and not force:
                id = id + '_' + datetime.datetime.now().strftime('%Y%m%d%H%M%S')

        # validate filename
        id = self.window.controller.presets.validate_filename(id)
        is_new = False
        if id not in self.window.core.presets.items:
            is_new = True
            self.window.core.presets.items[id] = self.window.core.presets.build()
        elif not force:
            self.window.ui.dialogs.confirm(
                type='preset_exists',
                id=id,
                msg=trans('confirm.preset.overwrite'),
            )
            return

        # check if at least one mode is selected
        is_mode = False
        for check in modes:
            if self.window.controller.config.get_value(
                parent_id=self.id,
                key=check,
                option=self.options[check],
            ):
                is_mode = True
                break
        if mode != MODE_AGENT and not is_mode:
            self.window.ui.dialogs.alert(
                trans('alert.preset.no_chat_completion')
            )
            return

        # assign data from fields to preset object in items
        self.assign_data(id)

        if is_new:
            # assign tmp avatar
            if self.tmp_avatar is not None:
                self.window.core.presets.items[id].ai_avatar = self.tmp_avatar
                self.tmp_avatar = None
        else:
            self.tmp_avatar = None

        # if agent, assign experts and select only agent mode
        curr_mode = self.window.core.config.get('mode')
        if curr_mode == MODE_AGENT:
            self.window.core.presets.items[id].mode = [MODE_AGENT]
        elif curr_mode == MODE_AGENT_LLAMA:
            self.window.core.presets.items[id].mode = [MODE_AGENT_LLAMA]
        elif curr_mode == MODE_AGENT_OPENAI:
            self.window.core.presets.items[id].mode = [MODE_AGENT_OPENAI]

        # apply changes to current active preset
        current = self.window.core.config.get('preset')
        if current is not None and current == id:
            self.to_current(self.window.core.presets.items[id])
            self.window.core.config.save()

        # update current uuid
        self.current = self.window.core.presets.items[id].uuid

        # save
        no_scroll = False
        if not is_new:
            no_scroll = True
        self.window.core.presets.save(id)
        self.window.controller.presets.refresh(no_scroll=no_scroll)

        # close dialog
        if close:
            self.window.ui.dialogs.close('editor.preset.presets')
        else:
            # update ID in field
            self.window.controller.config.apply_value(
                parent_id=self.id,
                key="filename",
                option=self.options["filename"],
                value=id,
            )
        self.window.update_status(trans('status.preset.saved'))

        # sort by name
        self.window.core.presets.sort_by_name()

        # switch to editing preset on save
        self.window.controller.presets.set(mode, id)
        self.window.controller.presets.select_model()

        # update presets list
        no_scroll = False
        if not is_new:
            no_scroll = True
        self.window.controller.presets.refresh(no_scroll=no_scroll)
        self.opened = False

    def assign_data(self, id: str):
        """
        Assign data from fields to preset

        :param id: preset id (filename)
        """
        data_dict = {}
        for key in self.options:
            # assigned separately
            if key == "tool.function":
                continue
            data_dict[key] = self.window.controller.config.get_value(
                parent_id=self.id,
                key=key,
                option=self.options[key],
            )
        if data_dict['name'] is None or data_dict['name'] == "":
            data_dict['name'] = id + " " + trans('preset.untitled')
        if data_dict['model'] == '_':
            data_dict['model'] = None

        preset = self.window.core.presets.items[id]
        preset.from_dict(data_dict)
        preset.filename = id
        preset.tools = {
            'function': [],  # functions are assigned separately (below)
        }

        # extra options
        self.append_extra_options(preset)

        # avatar update
        self.update_avatar_config(preset)

    def to_current(self, preset: PresetItem):
        """
        Update preset field from editor

        :param preset: preset item
        """
        self.window.core.config.set('ai_name', preset.ai_name)
        self.window.core.config.set('user_name', preset.user_name)
        self.window.core.config.set('prompt', preset.prompt)
        self.window.core.config.set('temperature', preset.temperature)

    def from_current(self):
        """Copy data from current active preset"""
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="ai_name",
            option=self.options["ai_name"],
            value=self.window.core.config.get('ai_name'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="user_name",
            option=self.options["user_name"],
            value=self.window.core.config.get('user_name'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="prompt",
            option=self.options["prompt"],
            value=self.window.core.config.get('prompt'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="temperature",
            option=self.options["temperature"],
            value=self.window.core.config.get('temperature'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="model",
            option=self.options["model"],
            value=self.window.core.config.get('model'),
        )

    def update_from_global(self, key: str, value: Any):
        """
        Update field from global config

        :param key: field key
        :param value: field value
        """
        preset_id = self.window.core.config.get('preset')
        if preset_id is not None and preset_id != "":
            if preset_id in self.window.core.presets.items:
                preset = self.window.core.presets.items[preset_id]
                if key == 'preset.user_name':
                    preset.user_name = value
                    self.window.core.config.set('user_name', preset.user_name)
                elif key == 'preset.ai_name':
                    preset.ai_name = value
                    self.window.core.config.set('ai_name', preset.ai_name)
                self.window.core.presets.save(preset_id)

    def upload_avatar(self, file_path: str):
        """
        Update avatar config for preset

        :param file_path: path to the avatar file
        """
        preset = self.window.core.presets.get_by_uuid(self.current)
        presets_dir = self.window.core.config.get_user_dir("presets")
        avatars_dir = os.path.join(presets_dir, "avatars")
        preset_name = "_" if preset is None else preset.filename
        if not os.path.exists(avatars_dir):
            os.makedirs(avatars_dir, exist_ok=True)
        file_ext = os.path.splitext(file_path)[1]
        store_name = preset_name + "_" + datetime.datetime.now().strftime('%Y%m%d%H%M%S') + file_ext
        avatar_path = os.path.join(avatars_dir, store_name)

        # copy avatar to avatars directory
        if os.path.exists(avatar_path):
            os.remove(avatar_path)
        if os.path.exists(file_path):
            shutil.copy(file_path, avatar_path)
            if preset:
                preset.ai_avatar = store_name
            else:
                self.tmp_avatar = store_name
            self.window.controller.config.apply_value(
                parent_id=self.id,
                key="ai_avatar",
                option=self.options["ai_avatar"],
                value=store_name,
            )
            self.window.ui.nodes['preset.editor.avatar'].load_avatar(avatar_path)
            self.window.ui.nodes['preset.editor.avatar'].enable_remove_button(True)
        return avatar_path

    def update_avatar_config(self, preset: PresetItem):
        """
        Update avatar config for preset

        :param preset: preset item
        """
        avatar_path = preset.ai_avatar
        if avatar_path:
            file_path = os.path.join(
                self.window.core.config.get_user_dir("presets"),
                "avatars",
                avatar_path,
            )
            if not os.path.exists(file_path):
                self.window.ui.nodes['preset.editor.avatar'].remove_avatar()
                print("Avatar file does not exist:", file_path)
                return
            self.window.ui.nodes['preset.editor.avatar'].load_avatar(file_path)
            self.window.ui.nodes['preset.editor.avatar'].enable_remove_button(True)
        else:
            self.window.ui.nodes['preset.editor.avatar'].remove_avatar()

    def remove_avatar(self, force: bool = False):
        """
        Remove avatar from preset editor

        :param force: force remove avatar
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='preset.avatar.delete',
                id="",
                msg=trans('confirm.preset.avatar.delete'),
            )
            return
        preset = self.window.core.presets.get_by_uuid(self.current)
        if preset:
            current = preset.ai_avatar
            if current:
                presets_dir = self.window.core.config.get_user_dir("presets")
                avatars_dir = os.path.join(presets_dir, "avatars")
                avatar_path = os.path.join(avatars_dir, current)
                if os.path.exists(avatar_path):
                    os.remove(avatar_path)
            preset.ai_avatar = ""
        else:
            self.tmp_avatar = None

        self.window.ui.nodes['preset.editor.avatar'].remove_avatar()
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="ai_avatar",
            option=self.options["ai_avatar"],
            value="",
        )

    def toggle_tab(self, name: str, show: bool = True):
        """
        Show experts tab

        :param name: name of the tab
        :param show: Show or hide experts tab
        """
        tabs = self.window.ui.tabs['preset.editor.tabs']
        idx = self.TAB_IDX[name]
        if tabs is not None:
            if show:
                tabs.setTabEnabled(idx, True)
                tabs.setTabVisible(idx, True)
                self.experts.update_tab()
            else:
                tabs.setTabEnabled(idx, False)
                tabs.setTabVisible(idx, False)
